from __future__ import annotations

import copy
import pickle
import weakref
from abc import ABCMeta, abstractmethod

import pytest

from ibis.common.bases import (
    Abstract,
    AbstractMeta,
    Comparable,
    Final,
    FrozenSlotted,
    Immutable,
    Singleton,
    Slotted,
)
from ibis.common.caching import WeakCache


def test_classes_are_based_on_abstract():
    assert issubclass(Comparable, Abstract)
    assert issubclass(Final, Abstract)
    assert issubclass(Immutable, Abstract)
    assert issubclass(Singleton, Abstract)


def test_abstract():
    class Foo(Abstract):
        @abstractmethod
        def foo(self):
            ...

        @property
        @abstractmethod
        def bar(self):
            ...

    assert not issubclass(type(Foo), ABCMeta)
    assert issubclass(type(Foo), AbstractMeta)
    assert Foo.__abstractmethods__ == frozenset({"foo", "bar"})

    with pytest.raises(TypeError, match="Can't instantiate abstract class .*Foo.*"):
        Foo()

    class Bar(Foo):
        def foo(self):
            return 1

        @property
        def bar(self):
            return 2

    bar = Bar()
    assert bar.foo() == 1
    assert bar.bar == 2
    assert isinstance(bar, Foo)
    assert isinstance(bar, Abstract)
    assert Bar.__abstractmethods__ == frozenset()


def test_immutable():
    class Foo(Immutable):
        __slots__ = ("a", "b")

        def __init__(self, a, b):
            object.__setattr__(self, "a", a)
            object.__setattr__(self, "b", b)

    foo = Foo(1, 2)
    assert foo.a == 1
    assert foo.b == 2
    with pytest.raises(AttributeError):
        foo.a = 2
    with pytest.raises(AttributeError):
        foo.b = 3

    assert copy.copy(foo) is foo
    assert copy.deepcopy(foo) is foo


class Node(Comparable):
    # override the default cache object
    __cache__ = WeakCache()
    __slots__ = ("name",)
    num_equal_calls = 0

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Node(name={self.name})"

    def __equals__(self, other):
        Node.num_equal_calls += 1
        return self.name == other.name


@pytest.fixture
def cache():
    Node.num_equal_calls = 0
    cache = Node.__cache__
    yield cache
    assert not cache


def pair(a, b):
    # for same ordering with comparable
    if id(a) < id(b):
        return (a, b)
    else:
        return (b, a)


def test_comparable_basic(cache):
    a = Node(name="a")
    b = Node(name="a")
    c = Node(name="a")
    assert a == b
    assert a == c
    del a
    del b
    del c


def test_comparable_caching(cache):
    a = Node(name="a")
    b = Node(name="b")
    c = Node(name="c")
    d = Node(name="d")
    e = Node(name="e")

    cache[pair(a, b)] = True
    cache[pair(a, c)] = False
    cache[pair(c, d)] = True
    cache[pair(b, d)] = False
    assert len(cache) == 4

    assert a == b
    assert a != c
    assert c == d
    assert b != d
    assert Node.num_equal_calls == 0

    # no cache hit
    assert pair(a, e) not in cache
    assert a != e
    assert Node.num_equal_calls == 1
    assert len(cache) == 5

    # run only once
    assert e != a
    assert Node.num_equal_calls == 1
    assert pair(a, e) in cache


def test_comparable_garbage_collection(cache):
    a = Node(name="a")
    b = Node(name="b")
    c = Node(name="c")
    d = Node(name="d")

    cache[pair(a, b)] = True
    cache[pair(a, c)] = False
    cache[pair(c, d)] = True
    cache[pair(b, d)] = False

    assert weakref.getweakrefcount(a) == 2
    del c
    assert weakref.getweakrefcount(a) == 1
    del b
    assert weakref.getweakrefcount(a) == 0


def test_comparable_cache_reuse(cache):
    nodes = [
        Node(name="a"),
        Node(name="b"),
        Node(name="c"),
        Node(name="d"),
        Node(name="e"),
    ]

    expected = 0
    for a, b in zip(nodes, nodes):
        a == a  # noqa: B015
        a == b  # noqa: B015
        b == a  # noqa: B015
        if a != b:
            expected += 1
        assert Node.num_equal_calls == expected

    assert len(cache) == expected

    # check that cache is evicted once nodes get collected
    del nodes
    assert len(cache) == 0

    a = Node(name="a")
    b = Node(name="a")
    assert a == b


class OneAndOnly(Singleton):
    __instances__ = weakref.WeakValueDictionary()


class DataType(Singleton):
    __slots__ = ("nullable",)
    __instances__ = weakref.WeakValueDictionary()

    def __init__(self, nullable=True):
        self.nullable = nullable


def test_singleton_basics():
    one = OneAndOnly()
    only = OneAndOnly()
    assert one is only

    assert len(OneAndOnly.__instances__) == 1
    key = (OneAndOnly, (), ())
    assert OneAndOnly.__instances__[key] is one


def test_singleton_lifetime() -> None:
    one = OneAndOnly()
    assert len(OneAndOnly.__instances__) == 1

    del one
    assert len(OneAndOnly.__instances__) == 0


def test_singleton_with_argument() -> None:
    dt1 = DataType(nullable=True)
    dt2 = DataType(nullable=False)
    dt3 = DataType(nullable=True)

    assert dt1 is dt3
    assert dt1 is not dt2
    assert len(DataType.__instances__) == 2

    del dt3
    assert len(DataType.__instances__) == 2
    del dt1
    assert len(DataType.__instances__) == 1
    del dt2
    assert len(DataType.__instances__) == 0


def test_final():
    class A(Final):
        pass

    with pytest.raises(TypeError, match="Cannot inherit from final class .*A.*"):

        class B(A):
            pass


class MyObj(Slotted):
    __slots__ = ("a", "b")


def test_slotted():
    obj = MyObj(a=1, b=2)
    assert obj.a == 1
    assert obj.b == 2
    assert obj.__fields__ == ("a", "b")
    assert obj.__slots__ == ("a", "b")
    with pytest.raises(AttributeError):
        obj.c = 3

    obj2 = MyObj(a=1, b=2)
    assert obj == obj2
    assert obj is not obj2

    obj3 = MyObj(a=1, b=3)
    assert obj != obj3

    assert pickle.loads(pickle.dumps(obj)) == obj

    with pytest.raises(KeyError):
        MyObj(a=1)


class MyObj2(MyObj):
    __slots__ = ("c",)


def test_slotted_inheritance():
    obj = MyObj2(a=1, b=2, c=3)
    assert obj.a == 1
    assert obj.b == 2
    assert obj.c == 3
    assert obj.__fields__ == ("a", "b", "c")
    assert obj.__slots__ == ("c",)
    with pytest.raises(AttributeError):
        obj.d = 4

    obj2 = MyObj2(a=1, b=2, c=3)
    assert obj == obj2
    assert obj is not obj2

    obj3 = MyObj2(a=1, b=2, c=4)
    assert obj != obj3
    assert pickle.loads(pickle.dumps(obj)) == obj

    with pytest.raises(KeyError):
        MyObj2(a=1, b=2)


class MyFrozenObj(FrozenSlotted):
    __slots__ = ("a", "b")


class MyFrozenObj2(MyFrozenObj):
    __slots__ = ("c", "d")


def test_frozen_slotted():
    obj = MyFrozenObj(a=1, b=2)

    assert obj.a == 1
    assert obj.b == 2
    assert obj.__fields__ == ("a", "b")
    assert obj.__slots__ == ("a", "b")
    with pytest.raises(AttributeError):
        obj.b = 3
    with pytest.raises(AttributeError):
        obj.c = 3

    obj2 = MyFrozenObj(a=1, b=2)
    assert obj == obj2
    assert obj is not obj2
    assert hash(obj) == hash(obj2)

    restored = pickle.loads(pickle.dumps(obj))
    assert restored == obj
    assert hash(restored) == hash(obj)

    with pytest.raises(KeyError):
        MyFrozenObj(a=1)


def test_frozen_slotted_inheritance():
    obj3 = MyFrozenObj2(a=1, b=2, c=3, d=4)
    assert obj3.__slots__ == ("c", "d")
    assert obj3.__fields__ == ("a", "b", "c", "d")
    assert pickle.loads(pickle.dumps(obj3)) == obj3



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import contextlib

from sqlglot.dialects import DuckDB

import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis.backends.base import BaseBackend
from ibis.expr.schema import Schema
from ibis.expr.tests.conftest import MOCK_TABLES


class MockBackend(BaseBackend):
    name = "mock"
    version = "1.0"
    current_database = "mockdb"
    dialect = DuckDB

    def __init__(self):
        super().__init__()
        self.executed_queries = []
        self.sql_query_schemas = {}

    def do_connect(self):
        pass

    def disconnect(self):
        pass

    def table(self, name, **kwargs):
        schema = self.get_schema(name)
        node = ops.DatabaseTable(source=self, name=name, schema=schema)
        return node.to_expr()

    def list_tables(self):
        return list(MOCK_TABLES)

    def list_databases(self):
        return ["mockdb"]

    def _to_sql(self, expr, **kwargs):
        import ibis

        return ibis.to_sql(expr, dialect="duckdb", **kwargs)

    def fetch_from_cursor(self, cursor, schema):
        pass

    def get_schema(self, name):
        name = name.replace("`", "")
        return Schema.from_tuples(MOCK_TABLES[name])

    def to_pyarrow(self, *_, **__):
        raise NotImplementedError(self.name)

    def execute(self, *_, **__):
        raise NotImplementedError(self.name)

    def compile(self, *_, **__):
        raise NotImplementedError(self.name)

    def create_table(self, *_, **__) -> ir.Table:
        raise NotImplementedError(self.name)

    def drop_table(self, *_, **__) -> ir.Table:
        raise NotImplementedError(self.name)

    def create_view(self, *_, **__) -> ir.Table:
        raise NotImplementedError(self.name)

    def drop_view(self, *_, **__) -> ir.Table:
        raise NotImplementedError(self.name)

    def _load_into_cache(self, *_):
        raise NotImplementedError(self.name)

    def _clean_up_cached_table(self, _):
        raise NotImplementedError(self.name)

    def _get_schema_using_query(self, query):
        return self.sql_query_schemas[query]

    def _get_sql_string_view_schema(self, name, table, query):
        return self.sql_query_schemas[query]

    @contextlib.contextmanager
    def set_query_schema(self, query, schema):
        self.sql_query_schemas[query] = schema
        yield
        self.sql_query_schemas.pop(query, None)



from __future__ import annotations

import collections.abc
from abc import abstractmethod
from typing import TYPE_CHECKING, Any
from weakref import WeakValueDictionary

from ibis.common.caching import WeakCache

if TYPE_CHECKING:
    from collections.abc import Mapping

    from typing_extensions import Self


class AbstractMeta(type):
    """Base metaclass for many of the ibis core classes.

    Enforce the subclasses to define a `__slots__` attribute and provide a
    `__create__` classmethod to change the instantiation behavior of the class.

    Support abstract methods without extending `abc.ABCMeta`. While it provides
    a reduced feature set compared to `abc.ABCMeta` (no way to register virtual
    subclasses) but avoids expensive instance checks by enforcing explicit
    subclassing.
    """

    __slots__ = ()

    def __new__(metacls, clsname, bases, dct, **kwargs):
        # enforce slot definitions
        dct.setdefault("__slots__", ())

        # construct the class object
        cls = super().__new__(metacls, clsname, bases, dct, **kwargs)

        # calculate abstract methods existing in the class
        abstracts = {
            name
            for name, value in dct.items()
            if getattr(value, "__isabstractmethod__", False)
        }
        for parent in bases:
            for name in getattr(parent, "__abstractmethods__", set()):
                value = getattr(cls, name, None)
                if getattr(value, "__isabstractmethod__", False):
                    abstracts.add(name)

        # set the abstract methods for the class
        cls.__abstractmethods__ = frozenset(abstracts)

        return cls

    def __call__(cls, *args, **kwargs):
        """Create a new instance of the class.

        The subclass may override the `__create__` classmethod to change the
        instantiation behavior. This is similar to overriding the `__new__`
        method, but without conditionally calling the `__init__` based on the
        return type.

        Parameters
        ----------
        args : tuple
            Positional arguments eventually passed to the `__init__` method.
        kwargs : dict
            Keyword arguments eventually passed to the `__init__` method.

        Returns
        -------
        The newly created instance of the class. No extra initialization

        """
        return cls.__create__(*args, **kwargs)


class Abstract(metaclass=AbstractMeta):
    """Base class for many of the ibis core classes, see `AbstractMeta`."""

    __slots__ = ("__weakref__",)
    __create__ = classmethod(type.__call__)  # type: ignore


class Immutable(Abstract):
    """Prohibit attribute assignment on the instance."""

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __setattr__(self, name: str, _: Any) -> None:
        raise AttributeError(
            f"Attribute {name!r} cannot be assigned to immutable instance of "
            f"type {type(self)}"
        )


class Singleton(Abstract):
    """Cache instances of the class based on instantiation arguments."""

    __instances__: Mapping[Any, Self] = WeakValueDictionary()

    @classmethod
    def __create__(cls, *args, **kwargs):
        key = (cls, args, tuple(kwargs.items()))
        try:
            return cls.__instances__[key]
        except KeyError:
            instance = super().__create__(*args, **kwargs)
            cls.__instances__[key] = instance
            return instance


class Final(Abstract):
    """Prohibit subclassing."""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.__init_subclass__ = cls.__prohibit_inheritance__

    @classmethod
    def __prohibit_inheritance__(cls, **kwargs):
        raise TypeError(f"Cannot inherit from final class {cls}")


@collections.abc.Hashable.register
class Hashable(Abstract):
    @abstractmethod
    def __hash__(self) -> int:
        ...


class Comparable(Abstract):
    """Enable quick equality comparisons.

    The subclasses must implement the `__equals__` method that returns a boolean
    value indicating whether the two instances are equal. This method is called
    only if the two instances are of the same type and the result is cached for
    future comparisons.

    Since the class holds a global cache of comparison results, it is important
    to make sure that the instances are not kept alive longer than necessary.
    This is done automatically by using weak references for the compared objects.
    """

    __cache__ = WeakCache()

    def __eq__(self, other) -> bool:
        try:
            return self.__cached_equals__(other)
        except TypeError:
            return NotImplemented

    @abstractmethod
    def __equals__(self, other) -> bool:
        ...

    def __cached_equals__(self, other) -> bool:
        if self is other:
            return True

        # type comparison should be cheap
        if type(self) is not type(other):
            return False

        # reduce space required for commutative operation
        if id(self) < id(other):
            key = (self, other)
        else:
            key = (other, self)

        try:
            result = self.__cache__[key]
        except KeyError:
            result = self.__equals__(other)
            self.__cache__[key] = result

        return result


class SlottedMeta(AbstractMeta):
    def __new__(metacls, clsname, bases, dct, **kwargs):
        fields = dct.get("__fields__", dct.get("__slots__", ()))
        inherited = (getattr(base, "__fields__", ()) for base in bases)
        dct["__fields__"] = sum(inherited, ()) + fields
        return super().__new__(metacls, clsname, bases, dct, **kwargs)


class Slotted(Abstract, metaclass=SlottedMeta):
    """A lightweight alternative to `ibis.common.grounds.Annotable`.

    The class is mostly used to reduce boilerplate code.
    """

    def __init__(self, **kwargs) -> None:
        for field in self.__fields__:
            object.__setattr__(self, field, kwargs[field])

    def __eq__(self, other) -> bool:
        if self is other:
            return True
        if type(self) is not type(other):
            return NotImplemented
        return all(getattr(self, n) == getattr(other, n) for n in self.__fields__)

    def __getstate__(self):
        return {k: getattr(self, k) for k in self.__fields__}

    def __setstate__(self, state):
        for name, value in state.items():
            object.__setattr__(self, name, value)

    def __repr__(self):
        fields = {k: getattr(self, k) for k in self.__fields__}
        fieldstring = ", ".join(f"{k}={v!r}" for k, v in fields.items())
        return f"{self.__class__.__name__}({fieldstring})"

    def __rich_repr__(self):
        for name in self.__fields__:
            yield name, getattr(self, name)


class FrozenSlotted(Slotted, Immutable, Hashable):
    """A lightweight alternative to `ibis.common.grounds.Concrete`.

    This class is used to create immutable dataclasses with slots and a precomputed
    hash value for quicker dictionary lookups.
    """

    __slots__ = ("__precomputed_hash__",)
    __fields__ = ()
    __precomputed_hash__: int

    def __init__(self, **kwargs) -> None:
        values = []
        for field in self.__fields__:
            values.append(value := kwargs[field])
            object.__setattr__(self, field, value)
        hashvalue = hash((self.__class__, tuple(values)))
        object.__setattr__(self, "__precomputed_hash__", hashvalue)

    def __setstate__(self, state):
        for name, value in state.items():
            object.__setattr__(self, name, value)
        hashvalue = hash((self.__class__, tuple(state.values())))
        object.__setattr__(self, "__precomputed_hash__", hashvalue)

    def __hash__(self) -> int:
        return self.__precomputed_hash__



from __future__ import annotations

import json
import os
import re
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from itertools import chain, repeat
from typing import TYPE_CHECKING, Any

import pytest
from requests import Session

import ibis
from ibis.backends.tests.base import ServiceBackendTest

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    import ibis.expr.types as ir

DRUID_URL = os.environ.get(
    "DRUID_URL", "druid://localhost:8082/druid/v2/sql?header=true"
)

# A different endpoint is required for DDL queries :(
BASE_URL = "http://localhost:8888/druid"

DDL_URL = f"{BASE_URL}/v2/sql/task/"
DDL_HEADERS = {"Content-Type": "application/json"}

REQUEST_INTERVAL = 0.5  # seconds


class DruidDataLoadError(Exception):
    pass


def wait_for_ingest(session: Session, *, datasource: str) -> bool:
    """Wait for datasources to be queryable."""
    # https://druid.apache.org/docs/latest/ingestion/faq.html#how-do-i-know-when-i-can-make-query-to-druid-after-submitting-batch-ingestion-task
    #
    # Steps 1 and 2: If we're in this function then the ingestion task has been
    # submitted (step 1) and the task is complete (step 2).
    #
    # Steps 3 and 4: Poll the segment loading by datasource API once with
    # forced metadata refresh (step 3) and then continue to do without forcing
    # until all segments for the data source are loaded (step 4)
    all_segments_loaded = False
    url = f"{BASE_URL}/coordinator/v1/datasources/{datasource}/loadstatus"
    force_refresh = chain(("true",), repeat("false"))
    while not all_segments_loaded:
        resp = session.get(url, params={"forceMetadataRefresh": next(force_refresh)})
        resp.raise_for_status()
        js = resp.json()

        # floating point comparison ¯\_(ツ)_/¯
        all_segments_loaded = js[datasource] == 100.0
        time.sleep(REQUEST_INTERVAL)
    return all_segments_loaded


def run_query(session: Session, query: str) -> None:
    """Run a data loading query."""
    resp = session.post(
        DDL_URL, data=json.dumps(dict(query=query)), headers=DDL_HEADERS
    )
    resp.raise_for_status()
    js = resp.json()

    task_id = js["taskId"]
    url = f"{BASE_URL}/indexer/v1/task/{task_id}/status"

    all_data_queryable = False

    while not all_data_queryable:
        resp = session.get(url)
        resp.raise_for_status()
        js = resp.json()

        status_blob = js["status"]
        status_string = status_blob["status"]
        if status_string == "SUCCESS":
            match = re.search(r'^REPLACE INTO "(?P<datasource>\w+)"', query)
            all_data_queryable = wait_for_ingest(
                session, datasource=match.groupdict()["datasource"]
            )
        elif status_string == "FAILED":
            raise DruidDataLoadError(status_blob["errorMsg"])
        time.sleep(REQUEST_INTERVAL)


class TestConf(ServiceBackendTest):
    # druid has the same rounding behavior as postgres
    check_dtype = False
    supports_window_operations = False
    returned_timestamp_unit = "s"
    supports_arrays = False
    supports_arrays_outside_of_select = supports_arrays
    native_bool = True
    supports_structs = False
    supports_json = False  # it does, but we haven't implemented it
    rounding_method = "half_to_even"
    service_name = "druid-middlemanager"
    deps = ("pydruid.db",)

    @property
    def functional_alltypes(self) -> ir.Table:
        t = self.connection.table("functional_alltypes")
        # The parquet loading for booleans appears to be broken in Druid, so
        # I'm using this as a workaround to make the data match what's on disk.
        return t.mutate(bool_col=1 - t.id % 2)

    @property
    def test_files(self) -> Iterable[Path]:
        return self.data_dir.joinpath("parquet").glob("*.parquet")

    def _load_data(self, **_: Any) -> None:
        """Load test data into a druid backend instance.

        Parameters
        ----------
        data_dir
            Location of testdata
        script_dir
            Location of scripts defining schemas
        """
        # run queries concurrently using threads; lots of time is spent on IO
        # making requests to check whether data loading is complete
        with Session() as session, ThreadPoolExecutor() as executor:
            for fut in as_completed(
                executor.submit(run_query, session, query) for query in self.ddl_script
            ):
                fut.result()

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw):
        return ibis.connect(DRUID_URL, **kw)


@pytest.fixture(scope="session")
def con(data_dir, tmp_path_factory, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection



from __future__ import annotations

import operator

import pandas as pd
import pandas.testing as tm
import pytest

import ibis


@pytest.fixture(scope="module")
def time_df1():
    return pd.DataFrame(
        {
            "time": pd.to_datetime([1, 2, 3, 4], unit="s"),
            "value": [1.1, 2.2, 3.3, 4.4],
            "group": ["a", "a", "a", "a"],
        }
    )


@pytest.fixture(scope="module")
def time_df2():
    return pd.DataFrame(
        {
            "time": pd.to_datetime([2, 4], unit="s"),
            "other_value": [1.2, 2.0],
            "group": ["a", "a"],
        }
    )


@pytest.fixture(scope="module")
def time_keyed_df1():
    return pd.DataFrame(
        {
            "time": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=6)
            ),
            "key": [1, 2, 3, 1, 2, 3],
            "value": [1.2, 1.4, 2.0, 4.0, 8.0, 16.0],
        }
    )


@pytest.fixture(scope="module")
def time_keyed_df2():
    return pd.DataFrame(
        {
            "time": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", freq="3D", periods=3)
            ),
            "key": [1, 2, 3],
            "other_value": [1.1, 1.2, 2.2],
        }
    )


@pytest.fixture(scope="module")
def time_left(time_df1):
    return ibis.memtable(time_df1)


@pytest.fixture(scope="module")
def time_right(time_df2):
    return ibis.memtable(time_df2)


@pytest.fixture(scope="module")
def time_keyed_left(time_keyed_df1):
    return ibis.memtable(time_keyed_df1)


@pytest.fixture(scope="module")
def time_keyed_right(time_keyed_df2):
    return ibis.memtable(time_keyed_df2)


@pytest.mark.parametrize(
    ("direction", "op"), [("backward", operator.ge), ("forward", operator.le)]
)
@pytest.mark.notyet(
    [
        "datafusion",
        "snowflake",
        "trino",
        "postgres",
        "mysql",
        "pyspark",
        "druid",
        "impala",
        "bigquery",
        "exasol",
        "oracle",
        "mssql",
        "sqlite",
        "risingwave",
        "flink",
    ]
)
def test_asof_join(con, time_left, time_right, time_df1, time_df2, direction, op):
    on = op(time_left["time"], time_right["time"])
    expr = time_left.asof_join(time_right, on=on, predicates="group")

    result = con.execute(expr)
    expected = pd.merge_asof(
        time_df1, time_df2, on="time", by="group", direction=direction
    )

    result = result.sort_values(["group", "time"]).reset_index(drop=True)
    expected = expected.sort_values(["group", "time"]).reset_index(drop=True)

    tm.assert_frame_equal(result[expected.columns], expected)
    with pytest.raises(AssertionError):
        tm.assert_series_equal(result["time"], result["time_right"])


@pytest.mark.parametrize(
    ("direction", "op"), [("backward", operator.ge), ("forward", operator.le)]
)
@pytest.mark.broken(
    ["clickhouse"], raises=AssertionError, reason="`time` is truncated to seconds"
)
@pytest.mark.notyet(
    [
        "datafusion",
        "snowflake",
        "trino",
        "postgres",
        "mysql",
        "pyspark",
        "druid",
        "impala",
        "bigquery",
        "exasol",
        "oracle",
        "mssql",
        "sqlite",
        "risingwave",
        "flink",
    ]
)
def test_keyed_asof_join_with_tolerance(
    con,
    time_keyed_left,
    time_keyed_right,
    time_keyed_df1,
    time_keyed_df2,
    direction,
    op,
):
    on = op(time_keyed_left["time"], time_keyed_right["time"])
    expr = time_keyed_left.asof_join(
        time_keyed_right, on=on, by="key", tolerance=ibis.interval(days=2)
    )

    result = con.execute(expr)
    expected = pd.merge_asof(
        time_keyed_df1,
        time_keyed_df2,
        on="time",
        by="key",
        tolerance=pd.Timedelta("2D"),
        direction=direction,
    )

    result = result.sort_values(["key", "time"]).reset_index(drop=True)
    expected = expected.sort_values(["key", "time"]).reset_index(drop=True)

    tm.assert_frame_equal(result[expected.columns], expected)
    with pytest.raises(AssertionError):
        tm.assert_series_equal(result["time"], result["time_right"])
    with pytest.raises(AssertionError):
        tm.assert_series_equal(result["key"], result["key_right"])



from __future__ import annotations

import string
from functools import partial, reduce

import sqlglot as sg
import sqlglot.expressions as sge
from public import public

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.base.sqlglot.compiler import NULL, STAR, SQLGlotCompiler
from ibis.backends.base.sqlglot.datatypes import MySQLType
from ibis.backends.base.sqlglot.dialects import MySQL
from ibis.backends.base.sqlglot.rewrites import (
    exclude_unsupported_window_frame_from_ops,
    exclude_unsupported_window_frame_from_rank,
    exclude_unsupported_window_frame_from_row_number,
    rewrite_empty_order_by_window,
    rewrite_first_to_first_value,
    rewrite_last_to_last_value,
    rewrite_sample_as_filter,
)
from ibis.common.patterns import replace
from ibis.expr.rewrites import p


@replace(p.Limit)
def rewrite_limit(_, **kwargs):
    """Rewrite limit for MySQL to include a large upper bound.

    From the MySQL docs @ https://dev.mysql.com/doc/refman/8.0/en/select.html

    > To retrieve all rows from a certain offset up to the end of the result
    > set, you can use some large number for the second parameter. This statement
    > retrieves all rows from the 96th row to the last:
    >
    > SELECT * FROM tbl LIMIT 95,18446744073709551615;
    """
    if _.n is None and _.offset is not None:
        some_large_number = (1 << 64) - 1
        return _.copy(n=some_large_number)
    return _


@public
class MySQLCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = MySQL
    type_mapper = MySQLType
    rewrites = (
        rewrite_limit,
        rewrite_sample_as_filter,
        rewrite_first_to_first_value,
        rewrite_last_to_last_value,
        exclude_unsupported_window_frame_from_ops,
        exclude_unsupported_window_frame_from_rank,
        exclude_unsupported_window_frame_from_row_number,
        rewrite_empty_order_by_window,
        *SQLGlotCompiler.rewrites,
    )

    @property
    def NAN(self):
        raise NotImplementedError("MySQL does not support NaN")

    @property
    def POS_INF(self):
        raise NotImplementedError("MySQL does not support Infinity")

    NEG_INF = POS_INF
    UNSUPPORTED_OPERATIONS = frozenset(
        (
            ops.ApproxMedian,
            ops.Arbitrary,
            ops.ArgMax,
            ops.ArgMin,
            ops.ArrayCollect,
            ops.Array,
            ops.ArrayFlatten,
            ops.ArrayMap,
            ops.Covariance,
            ops.First,
            ops.Last,
            ops.Levenshtein,
            ops.Median,
            ops.Mode,
            ops.MultiQuantile,
            ops.Quantile,
            ops.RegexReplace,
            ops.RegexSplit,
            ops.RowID,
            ops.StringSplit,
            ops.StructColumn,
            ops.TimestampBucket,
            ops.TimestampDelta,
            ops.Translate,
            ops.Unnest,
        )
    )

    SIMPLE_OPS = {
        ops.BitAnd: "bit_and",
        ops.BitOr: "bit_or",
        ops.BitXor: "bit_xor",
        ops.DayOfWeekName: "dayname",
        ops.Log10: "log10",
        ops.StringContains: "instr",
        ops.ExtractWeekOfYear: "weekofyear",
        ops.ExtractEpochSeconds: "unix_timestamp",
        ops.ExtractDayOfYear: "dayofyear",
        ops.Strftime: "date_format",
        ops.StringToTimestamp: "str_to_date",
        ops.Log2: "log2",
    }

    def _aggregate(self, funcname: str, *args, where):
        func = self.f[funcname]
        if where is not None:
            args = tuple(self.if_(where, arg, NULL) for arg in args)
        return func(*args)

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    def visit_Cast(self, op, *, arg, to):
        from_ = op.arg.dtype
        if (from_.is_json() or from_.is_string()) and to.is_json():
            # MariaDB does not support casting to JSON because it's an alias
            # for TEXT (except when casting of course!)
            return arg
        elif from_.is_integer() and to.is_interval():
            return self.visit_IntervalFromInteger(
                ops.IntervalFromInteger(op.arg, unit=to.unit), arg=arg, unit=to.unit
            )
        elif from_.is_integer() and to.is_timestamp():
            return self.f.from_unixtime(arg)
        return super().visit_Cast(op, arg=arg, to=to)

    def visit_TimestampDiff(self, op, *, left, right):
        return self.f.timestampdiff(
            sge.Var(this="SECOND"), right, left, dialect=self.dialect
        )

    def visit_DateDiff(self, op, *, left, right):
        return self.f.timestampdiff(
            sge.Var(this="DAY"), right, left, dialect=self.dialect
        )

    def visit_ApproxCountDistinct(self, op, *, arg, where):
        if where is not None:
            arg = self.if_(where, arg)
        return self.f.count(sge.Distinct(expressions=[arg]))

    def visit_CountStar(self, op, *, arg, where):
        if where is not None:
            return self.f.sum(self.cast(where, op.dtype))
        return self.f.count(STAR)

    def visit_CountDistinct(self, op, *, arg, where):
        if where is not None:
            arg = self.if_(where, arg)
        return self.f.count(sge.Distinct(expressions=[arg]))

    def visit_CountDistinctStar(self, op, *, arg, where):
        if where is not None:
            raise com.UnsupportedOperationError(
                "Filtered table count distinct is not supported in MySQL"
            )
        func = partial(sg.column, table=arg.alias_or_name, quoted=self.quoted)
        return self.f.count(
            sge.Distinct(expressions=list(map(func, op.arg.schema.keys())))
        )

    def visit_GroupConcat(self, op, *, arg, sep, where):
        if not isinstance(op.sep, ops.Literal):
            raise com.UnsupportedOperationError(
                "Only string literal separators are supported"
            )
        if where is not None:
            arg = self.if_(where, arg)
        return self.f.group_concat(arg, sep)

    def visit_DayOfWeekIndex(self, op, *, arg):
        return (self.f.dayofweek(arg) + 5) % 7

    def visit_Literal(self, op, *, value, dtype):
        # avoid casting NULL: the set of types allowed by MySQL and
        # MariaDB when casting is a strict subset of allowed types in other
        # contexts like CREATE TABLE
        if value is None:
            return NULL
        return super().visit_Literal(op, value=value, dtype=dtype)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_decimal() and not value.is_finite():
            raise com.UnsupportedOperationError(
                "MySQL does not support NaN or infinity"
            )
        elif dtype.is_binary():
            return self.f.unhex(value.hex())
        elif dtype.is_date():
            return self.f.date(value.isoformat())
        elif dtype.is_timestamp():
            return self.f.timestamp(value.isoformat())
        elif dtype.is_time():
            return self.f.maketime(
                value.hour, value.minute, value.second + value.microsecond / 1e6
            )
        elif dtype.is_array() or dtype.is_struct() or dtype.is_map():
            raise com.UnsupportedBackendType(
                "MySQL does not support arrays, structs or maps"
            )
        elif dtype.is_string():
            return sge.convert(value.replace("\\", "\\\\"))
        return None

    def visit_JSONGetItem(self, op, *, arg, index):
        if op.index.dtype.is_integer():
            path = self.f.concat("$[", self.cast(index, dt.string), "]")
        else:
            path = self.f.concat("$.", index)
        return self.f.json_extract(arg, path)

    def visit_DateFromYMD(self, op, *, year, month, day):
        return self.f.str_to_date(
            self.f.concat(
                self.f.lpad(year, 4, "0"),
                self.f.lpad(month, 2, "0"),
                self.f.lpad(day, 2, "0"),
            ),
            "%Y%m%d",
        )

    def visit_FindInSet(self, op, *, needle, values):
        return self.f.find_in_set(needle, self.f.concat_ws(",", values))

    def visit_EndsWith(self, op, *, arg, end):
        to = sge.DataType(this=sge.DataType.Type.BINARY)
        return self.f.right(arg, self.f.char_length(end)).eq(sge.Cast(this=end, to=to))

    def visit_StartsWith(self, op, *, arg, start):
        to = sge.DataType(this=sge.DataType.Type.BINARY)
        return self.f.left(arg, self.f.length(start)).eq(sge.Cast(this=start, to=to))

    def visit_RegexSearch(self, op, *, arg, pattern):
        return arg.rlike(pattern)

    def visit_RegexExtract(self, op, *, arg, pattern, index):
        extracted = self.f.regexp_substr(arg, pattern)
        return self.if_(
            arg.rlike(pattern),
            self.if_(
                index.eq(0),
                extracted,
                self.f.regexp_replace(
                    extracted, pattern, rf"\\{index.sql(self.dialect)}"
                ),
            ),
            NULL,
        )

    def visit_Equals(self, op, *, left, right):
        if op.left.dtype.is_string():
            assert op.right.dtype.is_string(), op.right.dtype
            to = sge.DataType(this=sge.DataType.Type.BINARY)
            return sge.Cast(this=left, to=to).eq(right)
        return super().visit_Equals(op, left=left, right=right)

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.instr(haystack, needle) > 0

    def visit_StringFind(self, op, *, arg, substr, start, end):
        if end is not None:
            raise NotImplementedError(
                "`end` argument is not implemented for MySQL `StringValue.find`"
            )
        substr = sge.Cast(this=substr, to=sge.DataType(this=sge.DataType.Type.BINARY))

        if start is not None:
            return self.f.locate(substr, arg, start + 1)
        return self.f.locate(substr, arg)

    def visit_LRStrip(self, op, *, arg, position):
        return reduce(
            lambda arg, char: self.f.trim(this=arg, position=position, expression=char),
            map(
                partial(self.cast, to=dt.string),
                map(self.f.unhex, map(self.f.hex, string.whitespace.encode())),
            ),
            arg,
        )

    def visit_DateTimestampTruncate(self, op, *, arg, unit):
        truncate_formats = {
            "s": "%Y-%m-%d %H:%i:%s",
            "m": "%Y-%m-%d %H:%i:00",
            "h": "%Y-%m-%d %H:00:00",
            "D": "%Y-%m-%d",
            # 'W': 'week',
            "M": "%Y-%m-01",
            "Y": "%Y-01-01",
        }
        if (format := truncate_formats.get(unit.short)) is None:
            raise com.UnsupportedOperationError(f"Unsupported truncate unit {op.unit}")
        return self.f.date_format(arg, format)

    visit_DateTruncate = visit_TimestampTruncate = visit_DateTimestampTruncate

    def visit_DateTimeDelta(self, op, *, left, right, part):
        return self.f.timestampdiff(
            sge.Var(this=part.this), right, left, dialect=self.dialect
        )

    visit_TimeDelta = visit_DateDelta = visit_DateTimeDelta

    def visit_ExtractMillisecond(self, op, *, arg):
        return self.f.floor(self.f.extract(sge.Var(this="microsecond"), arg) / 1_000)

    def visit_ExtractMicrosecond(self, op, *, arg):
        return self.f.floor(self.f.extract(sge.Var(this="microsecond"), arg))

    def visit_Strip(self, op, *, arg):
        return self.visit_LRStrip(op, arg=arg, position="BOTH")

    def visit_LStrip(self, op, *, arg):
        return self.visit_LRStrip(op, arg=arg, position="LEADING")

    def visit_RStrip(self, op, *, arg):
        return self.visit_LRStrip(op, arg=arg, position="TRAILING")

    def visit_IntervalFromInteger(self, op, *, arg, unit):
        return sge.Interval(this=arg, unit=sge.convert(op.resolution.upper()))

    def visit_TimestampAdd(self, op, *, left, right):
        if op.right.dtype.unit.short == "ms":
            right = sge.Interval(
                this=right.this * 1_000, unit=sge.Var(this="MICROSECOND")
            )
        return self.f.date_add(left, right, dialect=self.dialect)



from __future__ import annotations

import contextlib
from ipaddress import IPv4Address, IPv6Address

import pytest
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
from ibis.backends.tests.errors import ExaQueryError

MACADDR_BACKEND_TYPE = {
    "bigquery": "STRING",
    "clickhouse": "String",
    "duckdb": "VARCHAR",
    "snowflake": "VARCHAR",
    "sqlite": "text",
    "trino": "varchar(17)",
    "impala": "STRING",
    "postgres": "text",
    "risingwave": "text",
    "flink": "CHAR(17) NOT NULL",
}


@pytest.mark.notimpl(["polars"], raises=NotImplementedError)
def test_macaddr_literal(con, backend):
    test_macaddr = "00:00:0A:BB:28:FC"
    expr = ibis.literal(test_macaddr, type=dt.macaddr)
    result = con.execute(expr)
    assert result == test_macaddr

    with contextlib.suppress(com.OperationNotDefinedError):
        backend_name = backend.name()
        assert con.execute(expr.typeof()) == MACADDR_BACKEND_TYPE[backend_name]


@pytest.mark.parametrize(
    ("test_value", "expected_values", "expected_types"),
    [
        param(
            "127.0.0.1",
            {
                "bigquery": "127.0.0.1",
                "clickhouse": IPv4Address("127.0.0.1"),
                "duckdb": "127.0.0.1",
                "snowflake": "127.0.0.1",
                "sqlite": "127.0.0.1",
                "trino": "127.0.0.1",
                "impala": "127.0.0.1",
                "postgres": "127.0.0.1",
                "risingwave": "127.0.0.1",
                "pandas": "127.0.0.1",
                "pyspark": "127.0.0.1",
                "mysql": "127.0.0.1",
                "dask": "127.0.0.1",
                "mssql": "127.0.0.1",
                "datafusion": "127.0.0.1",
                "flink": "127.0.0.1",
            },
            {
                "bigquery": "STRING",
                "clickhouse": "IPv4",
                "duckdb": "VARCHAR",
                "snowflake": "VARCHAR",
                "sqlite": "text",
                "trino": "varchar(9)",
                "impala": "STRING",
                "postgres": "text",
                "risingwave": "text",
                "flink": "CHAR(9) NOT NULL",
            },
            id="ipv4",
        ),
        param(
            "2001:db8::1",
            {
                "bigquery": "2001:db8::1",
                "clickhouse": IPv6Address("2001:db8::1"),
                "duckdb": "2001:db8::1",
                "snowflake": "2001:db8::1",
                "sqlite": "2001:db8::1",
                "trino": "2001:db8::1",
                "impala": "2001:db8::1",
                "postgres": "2001:db8::1",
                "risingwave": "2001:db8::1",
                "pandas": "2001:db8::1",
                "pyspark": "2001:db8::1",
                "mysql": "2001:db8::1",
                "dask": "2001:db8::1",
                "mssql": "2001:db8::1",
                "datafusion": "2001:db8::1",
                "flink": "2001:db8::1",
            },
            {
                "bigquery": "STRING",
                "clickhouse": "IPv6",
                "duckdb": "VARCHAR",
                "snowflake": "VARCHAR",
                "sqlite": "text",
                "trino": "varchar(11)",
                "impala": "STRING",
                "postgres": "text",
                "risingwave": "text",
                "flink": "CHAR(11) NOT NULL",
            },
            id="ipv6",
        ),
    ],
)
@pytest.mark.notimpl(["polars"], raises=NotImplementedError)
@pytest.mark.notimpl(["druid", "oracle"], raises=KeyError)
@pytest.mark.notimpl(["exasol"], raises=(ExaQueryError, KeyError))
def test_inet_literal(con, backend, test_value, expected_values, expected_types):
    backend_name = backend.name()
    expr = ibis.literal(test_value, type=dt.inet)
    result = con.execute(expr)

    assert result == expected_values[backend_name]

    with contextlib.suppress(com.OperationNotDefinedError):
        assert con.execute(expr.typeof()) == expected_types[backend_name]



from __future__ import annotations

import polars as pl
import pytest
from pytest import param

import ibis.expr.datatypes as dt
from ibis.backends.polars.datatypes import dtype_from_polars, dtype_to_polars


@pytest.mark.parametrize(
    ("ibis_dtype", "polars_type"),
    [
        param(dt.bool, pl.Boolean, id="bool"),
        param(dt.null, pl.Null, id="null"),
        param(dt.Array(dt.string), pl.List(pl.Utf8), id="array_string"),
        param(dt.string, pl.Utf8, id="string"),
        param(dt.binary, pl.Binary, id="binary"),
        param(dt.date, pl.Date, id="date"),
        param(dt.time, pl.Time, id="time"),
        param(dt.int8, pl.Int8, id="int8"),
        param(dt.int16, pl.Int16, id="int16"),
        param(dt.int32, pl.Int32, id="int32"),
        param(dt.int64, pl.Int64, id="int64"),
        param(dt.uint8, pl.UInt8, id="uint8"),
        param(dt.uint16, pl.UInt16, id="uint16"),
        param(dt.uint32, pl.UInt32, id="uint32"),
        param(dt.uint64, pl.UInt64, id="uint64"),
        param(dt.float32, pl.Float32, id="float32"),
        param(dt.float64, pl.Float64, id="float64"),
        param(dt.timestamp, pl.Datetime("ns", time_zone=None), id="timestamp"),
        param(
            dt.Timestamp("UTC"), pl.Datetime("ns", time_zone="UTC"), id="timestamp_tz"
        ),
        param(dt.Interval(unit="ms"), pl.Duration("ms"), id="interval_ms"),
        param(dt.Interval(unit="us"), pl.Duration("us"), id="interval_us"),
        param(dt.Interval(unit="ns"), pl.Duration("ns"), id="interval_ns"),
        param(
            dt.Struct(
                dict(a=dt.string, b=dt.Array(dt.Array(dt.Struct(dict(c=dt.float64)))))
            ),
            pl.Struct(
                [
                    pl.Field("a", pl.Utf8),
                    pl.Field(
                        "b", pl.List(pl.List(pl.Struct([pl.Field("c", pl.Float64)])))
                    ),
                ]
            ),
            id="struct",
        ),
    ],
)
def test_to_from_ibis_type(ibis_dtype, polars_type):
    assert dtype_to_polars(ibis_dtype) == polars_type
    assert dtype_from_polars(polars_type) == ibis_dtype


def test_categorical():
    assert dtype_from_polars(pl.Categorical()) == dt.string



from __future__ import annotations

import copy
import pickle
import sys
import weakref
from abc import ABCMeta
from typing import TYPE_CHECKING, Callable, Generic, Optional, TypeVar, Union

import pytest

from ibis.common.annotations import (
    Argument,
    Parameter,
    Signature,
    ValidationError,
    argument,
    attribute,
    optional,
    varargs,
    varkwargs,
)
from ibis.common.collections import Mapping, Sequence
from ibis.common.grounds import (
    Abstract,
    Annotable,
    AnnotableMeta,
    Comparable,
    Concrete,
    Immutable,
    Singleton,
)
from ibis.common.patterns import (
    Any,
    As,
    CoercedTo,
    Coercible,
    InstanceOf,
    Option,
    Pattern,
    TupleOf,
)
from ibis.tests.util import assert_pickle_roundtrip

if TYPE_CHECKING:
    from typing_extensions import Self

is_any = InstanceOf(object)
is_bool = InstanceOf(bool)
is_float = InstanceOf(float)
is_int = InstanceOf(int)
is_str = InstanceOf(str)
is_list = InstanceOf(list)


class Op(Annotable):
    pass


class Value(Op):
    arg = InstanceOf(object)


class StringOp(Value):
    arg = InstanceOf(str)


class BetweenSimple(Annotable):
    value = is_int
    lower = optional(is_int, default=0)
    upper = optional(is_int, default=None)


class BetweenWithExtra(Annotable):
    extra = attribute(is_int)
    value = is_int
    lower = optional(is_int, default=0)
    upper = optional(is_int, default=None)


class BetweenWithCalculated(Concrete):
    value = is_int
    lower = optional(is_int, default=0)
    upper = optional(is_int, default=None)

    @attribute
    def calculated(self):
        return self.value + self.lower


class VariadicArgs(Concrete):
    args = varargs(is_int)


class VariadicKeywords(Concrete):
    kwargs = varkwargs(is_int)


class VariadicArgsAndKeywords(Concrete):
    args = varargs(is_int)
    kwargs = varkwargs(is_int)


T = TypeVar("T", covariant=True)
K = TypeVar("K", covariant=True)
V = TypeVar("V", covariant=True)


class List(Concrete, Sequence[T], Coercible):
    @classmethod
    def __coerce__(self, values, T=None):
        values = tuple(values)
        if values:
            head, *rest = values
            return ConsList(head, rest)
        else:
            return EmptyList()


class EmptyList(List[T]):
    def __getitem__(self, key):
        raise IndexError(key)

    def __len__(self):
        return 0


class ConsList(List[T]):
    head: T
    rest: List[T]

    def __getitem__(self, key):
        if key == 0:
            return self.head
        else:
            return self.rest[key - 1]

    def __len__(self):
        return len(self.rest) + 1


class Map(Concrete, Mapping[K, V], Coercible):
    @classmethod
    def __coerce__(self, pairs, K=None, V=None):
        pairs = dict(pairs)
        if pairs:
            head_key = next(iter(pairs))
            head_value = pairs.pop(head_key)
            rest = pairs
            return ConsMap((head_key, head_value), rest)
        else:
            return EmptyMap()


class EmptyMap(Map[K, V]):
    def __getitem__(self, key):
        raise KeyError(key)

    def __iter__(self):
        return iter(())

    def __len__(self):
        return 0


class ConsMap(Map[K, V]):
    head: tuple[K, V]
    rest: Map[K, V]

    def __getitem__(self, key):
        if key == self.head[0]:
            return self.head[1]
        else:
            return self.rest[key]

    def __iter__(self):
        yield self.head[0]
        yield from self.rest

    def __len__(self):
        return len(self.rest) + 1


class Integer(int, Coercible):
    @classmethod
    def __coerce__(cls, value):
        return Integer(value)


class Float(float, Coercible):
    @classmethod
    def __coerce__(cls, value):
        return Float(value)


class MyExpr(Concrete):
    a: Integer
    b: List[Float]
    c: Map[str, Integer]


class MyInt(int, Coercible):
    @classmethod
    def __coerce__(cls, value):
        return cls(value)


class MyFloat(float, Coercible):
    @classmethod
    def __coerce__(cls, value):
        return cls(value)


J = TypeVar("J", bound=MyInt, covariant=True)
F = TypeVar("F", bound=MyFloat, covariant=True)
N = TypeVar("N", bound=Union[MyInt, MyFloat], covariant=True)


class MyValue(Annotable, Generic[J, F]):
    integer: J
    floating: F
    numeric: N


def test_annotable():
    class Between(BetweenSimple):
        pass

    assert not issubclass(type(Between), ABCMeta)
    assert type(Between) is AnnotableMeta

    argnames = ("value", "lower", "upper")
    signature = BetweenSimple.__signature__
    assert isinstance(signature, Signature)
    assert tuple(signature.parameters.keys()) == argnames
    assert BetweenSimple.__slots__ == argnames

    obj = BetweenSimple(10, lower=2)
    assert obj.value == 10
    assert obj.lower == 2
    assert obj.upper is None
    assert obj.__argnames__ == argnames
    assert obj.__slots__ == ("value", "lower", "upper")
    assert not hasattr(obj, "__dict__")
    assert obj.__module__ == __name__
    assert type(obj).__qualname__ == "BetweenSimple"

    # test that a child without additional arguments doesn't have __dict__
    obj = Between(10, lower=2)
    assert obj.__slots__ == tuple()
    assert not hasattr(obj, "__dict__")
    assert obj.__module__ == __name__
    assert type(obj).__qualname__ == "test_annotable.<locals>.Between"

    copied = copy.copy(obj)
    assert obj == copied
    assert obj is not copied

    copied = obj.copy()
    assert obj == copied
    assert obj is not copied

    obj2 = Between(10, lower=8)
    assert obj.copy(lower=8) == obj2


def test_annotable_with_bound_typevars_properly_coerce_values():
    v = MyValue(1.1, 2.2, 3.3)
    assert isinstance(v.integer, MyInt)
    assert v.integer == 1
    assert isinstance(v.floating, MyFloat)
    assert v.floating == 2.2
    assert isinstance(v.numeric, MyInt)
    assert v.numeric == 3


def test_annotable_with_additional_attributes():
    a = BetweenWithExtra(10, lower=2)
    b = BetweenWithExtra(10, lower=2)
    assert a == b
    assert a is not b

    a.extra = 1
    assert a.extra == 1
    assert a != b

    assert a == pickle.loads(pickle.dumps(a))


def test_annotable_is_mutable_by_default():
    # TODO(kszucs): more exhaustive testing of mutability, e.g. setting
    # optional value to None doesn't set to the default value
    class Op(Annotable):
        __slots__ = ("custom",)

        a = is_int
        b = is_int

    p = Op(1, 2)
    assert p.a == 1
    p.a = 3
    assert p.a == 3
    assert p == Op(a=3, b=2)

    # test that non-annotable attributes can be set as well
    p.custom = 1
    assert p.custom == 1


def test_annotable_with_type_annotations() -> None:
    # TODO(kszucs): bar: str = None  # should raise
    class Op1(Annotable):
        foo: int
        bar: str = ""

    p = Op1(1)
    assert p.foo == 1
    assert not p.bar

    class Op2(Annotable):
        bar: str = None

    with pytest.raises(ValidationError):
        Op2()


class RecursiveNode(Annotable):
    child: Optional[Self] = None


def test_annotable_with_self_typehint() -> None:
    node = RecursiveNode(RecursiveNode(RecursiveNode(None)))
    assert isinstance(node, RecursiveNode)
    assert isinstance(node.child, RecursiveNode)
    assert isinstance(node.child.child, RecursiveNode)
    assert node.child.child.child is None

    with pytest.raises(ValidationError):
        RecursiveNode(1)


def test_annotable_with_recursive_generic_type_annotations():
    # testing cons list
    pattern = Pattern.from_typehint(List[Integer])
    values = ["1", 2.0, 3]
    result = pattern.match(values, {})
    expected = ConsList(1, ConsList(2, ConsList(3, EmptyList())))
    assert result == expected
    assert result[0] == 1
    assert result[1] == 2
    assert result[2] == 3
    assert len(result) == 3
    with pytest.raises(IndexError):
        result[3]

    # testing cons map
    pattern = Pattern.from_typehint(Map[Integer, Float])
    values = {"1": 2, 3: "4.0", 5: 6.0}
    result = pattern.match(values, {})
    expected = ConsMap((1, 2.0), ConsMap((3, 4.0), ConsMap((5, 6.0), EmptyMap())))
    assert result == expected
    assert result[1] == 2.0
    assert result[3] == 4.0
    assert result[5] == 6.0
    assert len(result) == 3
    with pytest.raises(KeyError):
        result[7]

    # testing both encapsulated in a class
    expr = MyExpr(a="1", b=["2.0", 3, True], c={"a": "1", "b": 2, "c": 3.0})
    assert expr.a == 1
    assert expr.b == ConsList(2.0, ConsList(3.0, ConsList(1.0, EmptyList())))
    assert expr.c == ConsMap(("a", 1), ConsMap(("b", 2), ConsMap(("c", 3), EmptyMap())))


def test_composition_of_annotable_and_immutable():
    class AnnImm(Annotable, Immutable):
        value = is_int
        lower = optional(is_int, default=0)
        upper = optional(is_int, default=None)

    class ImmAnn(Immutable, Annotable):
        # this is the preferable method resolution order
        value = is_int
        lower = optional(is_int, default=0)
        upper = optional(is_int, default=None)

    obj = AnnImm(3, lower=0, upper=4)
    with pytest.raises(AttributeError):
        obj.value = 1

    obj = ImmAnn(3, lower=0, upper=4)
    with pytest.raises(AttributeError):
        obj.value = 1


def test_composition_of_annotable_and_comparable():
    class Between(Comparable, Annotable):
        value = is_int
        lower = optional(is_int, default=0)
        upper = optional(is_int, default=None)

        def __equals__(self, other):
            return (
                self.value == other.value
                and self.lower == other.lower
                and self.upper == other.upper
            )

    a = Between(3, lower=0, upper=4)
    b = Between(3, lower=0, upper=4)
    c = Between(2, lower=0, upper=4)

    assert Between.__eq__ is Comparable.__eq__
    assert a == b
    assert b == a
    assert a != c
    assert c != a
    assert a.__equals__(b)
    assert a.__cached_equals__(b)
    assert not a.__equals__(c)
    assert not a.__cached_equals__(c)


def test_maintain_definition_order():
    class Between(Annotable):
        value = is_int
        lower = optional(is_int, default=0)
        upper = optional(is_int, default=None)

    param_names = list(Between.__signature__.parameters.keys())
    assert param_names == ["value", "lower", "upper"]


def test_signature_inheritance():
    class IntBinop(Annotable):
        left = is_int
        right = is_int

    class FloatAddRhs(IntBinop):
        right = is_float

    class FloatAddClip(FloatAddRhs):
        left = is_float
        clip_lower = optional(is_int, default=0)
        clip_upper = optional(is_int, default=10)

    class IntAddClip(FloatAddClip, IntBinop):
        pass

    assert IntBinop.__signature__ == Signature(
        [
            Parameter("left", annotation=Argument(is_int)),
            Parameter("right", annotation=Argument(is_int)),
        ]
    )

    assert FloatAddRhs.__signature__ == Signature(
        [
            Parameter("left", annotation=Argument(is_int)),
            Parameter("right", annotation=Argument(is_float)),
        ]
    )

    assert FloatAddClip.__signature__ == Signature(
        [
            Parameter("left", annotation=Argument(is_float)),
            Parameter("right", annotation=Argument(is_float)),
            Parameter("clip_lower", annotation=optional(is_int, default=0)),
            Parameter("clip_upper", annotation=optional(is_int, default=10)),
        ]
    )

    assert IntAddClip.__signature__ == Signature(
        [
            Parameter("left", annotation=Argument(is_int)),
            Parameter("right", annotation=Argument(is_int)),
            Parameter("clip_lower", annotation=optional(is_int, default=0)),
            Parameter("clip_upper", annotation=optional(is_int, default=10)),
        ]
    )


def test_positional_argument_reordering():
    class Farm(Annotable):
        ducks = is_int
        donkeys = is_int
        horses = is_int
        goats = is_int
        chickens = is_int

    class NoHooves(Farm):
        horses = optional(is_int, default=0)
        goats = optional(is_int, default=0)
        donkeys = optional(is_int, default=0)

    f1 = Farm(1, 2, 3, 4, 5)
    f2 = Farm(1, 2, goats=4, chickens=5, horses=3)
    f3 = Farm(1, 0, 0, 0, 100)
    assert f1 == f2
    assert f1 != f3

    g1 = NoHooves(1, 2, donkeys=-1)
    assert g1.ducks == 1
    assert g1.chickens == 2
    assert g1.donkeys == -1
    assert g1.horses == 0
    assert g1.goats == 0


def test_keyword_argument_reordering():
    class Alpha(Annotable):
        a = is_int
        b = is_int

    class Beta(Alpha):
        c = is_int
        d = optional(is_int, default=0)
        e = is_int

    obj = Beta(1, 2, 3, 4)
    assert obj.a == 1
    assert obj.b == 2
    assert obj.c == 3
    assert obj.e == 4
    assert obj.d == 0

    obj = Beta(1, 2, 3, 4, 5)
    assert obj.d == 5
    assert obj.e == 4


def test_variadic_argument_reordering():
    class Test(Annotable):
        a = is_int
        b = is_int
        args = varargs(is_int)

    class Test2(Test):
        c = is_int
        args = varargs(is_int)

    with pytest.raises(ValidationError, match="missing a required argument: 'c'"):
        Test2(1, 2)

    a = Test2(1, 2, 3)
    assert a.a == 1
    assert a.b == 2
    assert a.c == 3
    assert a.args == ()

    b = Test2(*range(5))
    assert b.a == 0
    assert b.b == 1
    assert b.c == 2
    assert b.args == (3, 4)

    msg = "only one variadic \\*args parameter is allowed"
    with pytest.raises(TypeError, match=msg):

        class Test3(Test):
            another_args = varargs(is_int)


def test_variadic_keyword_argument_reordering():
    class Test(Annotable):
        a = is_int
        b = is_int
        options = varkwargs(is_int)

    class Test2(Test):
        c = is_int
        options = varkwargs(is_int)

    with pytest.raises(ValidationError, match="missing a required argument: 'c'"):
        Test2(1, 2)

    a = Test2(1, 2, c=3)
    assert a.a == 1
    assert a.b == 2
    assert a.c == 3
    assert a.options == {}

    b = Test2(1, 2, c=3, d=4, e=5)
    assert b.a == 1
    assert b.b == 2
    assert b.c == 3
    assert b.options == {"d": 4, "e": 5}

    msg = "only one variadic \\*\\*kwargs parameter is allowed"
    with pytest.raises(TypeError, match=msg):

        class Test3(Test):
            another_options = varkwargs(is_int)


def test_variadic_argument():
    class Test(Annotable):
        a = is_int
        b = is_int
        args = varargs(is_int)

    assert Test(1, 2).args == ()
    assert Test(1, 2, 3).args == (3,)
    assert Test(1, 2, 3, 4, 5).args == (3, 4, 5)


def test_variadic_keyword_argument():
    class Test(Annotable):
        first = is_int
        second = is_int
        options = varkwargs(is_int)

    assert Test(1, 2).options == {}
    assert Test(1, 2, a=3).options == {"a": 3}
    assert Test(1, 2, a=3, b=4, c=5).options == {"a": 3, "b": 4, "c": 5}


def test_copy_with_variadic_argument():
    class Foo(Annotable):
        a = is_int
        b = is_int
        args = varargs(is_int)

    class Bar(Concrete):
        a = is_int
        b = is_int
        args = varargs(is_int)

    for t in [Foo(1, 2, 3, 4, 5), Bar(1, 2, 3, 4, 5)]:
        assert t.a == 1
        assert t.b == 2
        assert t.args == (3, 4, 5)

        u = t.copy(a=6, args=(8, 9, 10))
        assert u.a == 6
        assert u.b == 2
        assert u.args == (8, 9, 10)


def test_concrete_copy_with_unknown_argument_raise():
    class Bar(Concrete):
        a = is_int
        b = is_int

    t = Bar(1, 2)
    assert t.a == 1
    assert t.b == 2

    with pytest.raises(AttributeError, match="Unexpected arguments"):
        t.copy(c=3, d=4)


def test_concrete_pickling_variadic_arguments():
    v = VariadicArgs(1, 2, 3, 4, 5)
    assert v.args == (1, 2, 3, 4, 5)
    assert_pickle_roundtrip(v)

    v = VariadicKeywords(a=3, b=4, c=5)
    assert v.kwargs == {"a": 3, "b": 4, "c": 5}
    assert_pickle_roundtrip(v)

    v = VariadicArgsAndKeywords(1, 2, 3, 4, 5, a=3, b=4, c=5)
    assert v.args == (1, 2, 3, 4, 5)
    assert v.kwargs == {"a": 3, "b": 4, "c": 5}
    assert_pickle_roundtrip(v)


def test_dont_copy_default_argument():
    default = tuple()

    class Op(Annotable):
        arg = optional(InstanceOf(tuple), default=default)

    op = Op()
    assert op.arg is default


def test_copy_mutable_with_default_attribute():
    class Test(Annotable):
        a = attribute(InstanceOf(dict), default={})
        b = argument(InstanceOf(str))  # required argument

        @attribute
        def c(self):
            return self.b.upper()

    t = Test("t")
    assert t.a == {}
    assert t.b == "t"
    assert t.c == "T"

    with pytest.raises(ValidationError):
        t.a = 1
    t.a = {"map": "ping"}
    assert t.a == {"map": "ping"}

    assert t.copy() == t

    u = t.copy(b="u")
    assert u.b == "u"
    assert u.c == "T"
    assert u.a == {"map": "ping"}

    x = t.copy(a={"emp": "ty"})
    assert x.a == {"emp": "ty"}
    assert x.b == "t"


def test_slots_are_inherited_and_overridable():
    class Op(Annotable):
        __slots__ = ("_cache",)  # first definition
        arg = Any()

    class StringOp(Op):
        arg = CoercedTo(str)  # new overridden slot

    class StringSplit(StringOp):
        sep = CoercedTo(str)  # new slot

    class StringJoin(StringOp):
        __slots__ = ("_memoize",)  # new slot
        sep = CoercedTo(str)  # new overridden slot

    assert Op.__slots__ == ("_cache", "arg")
    assert StringOp.__slots__ == ("arg",)
    assert StringSplit.__slots__ == ("sep",)
    assert StringJoin.__slots__ == ("_memoize", "sep")


def test_multiple_inheritance():
    # multiple inheritance is allowed only if one of the parents has non-empty
    # __slots__ definition, otherwise python will raise lay-out conflict

    class Op(Annotable):
        __slots__ = ("_hash",)

    class Value(Annotable):
        arg = InstanceOf(object)

    class Reduction(Value):
        pass

    class UDF(Value):
        func = InstanceOf(Callable)

    class UDAF(UDF, Reduction):
        arity = is_int

    class A(Annotable):
        a = is_int

    class B(Annotable):
        b = is_int

    msg = "multiple bases have instance lay-out conflict"
    with pytest.raises(TypeError, match=msg):

        class AB(A, B):
            ab = is_int

    assert UDAF.__slots__ == ("arity",)
    strlen = UDAF(arg=2, func=lambda value: len(str(value)), arity=1)
    assert strlen.arg == 2
    assert strlen.arity == 1


@pytest.mark.parametrize(
    "obj",
    [
        StringOp("something"),
        StringOp(arg="something"),
    ],
)
def test_pickling_support(obj):
    assert_pickle_roundtrip(obj)


def test_multiple_inheritance_argument_order():
    class Value(Annotable):
        arg = is_any

    class VersionedOp(Value):
        version = is_int

    class Reduction(Annotable):
        pass

    class Sum(VersionedOp, Reduction):
        where = optional(is_bool, default=False)

    assert tuple(Sum.__signature__.parameters.keys()) == ("arg", "version", "where")


def test_multiple_inheritance_optional_argument_order():
    class Value(Annotable):
        pass

    class ConditionalOp(Annotable):
        where = optional(is_bool, default=False)

    class Between(Value, ConditionalOp):
        min = is_int
        max = is_int
        how = optional(is_str, default="strict")

    assert tuple(Between.__signature__.parameters.keys()) == (
        "min",
        "max",
        "how",
        "where",
    )


def test_immutability():
    class Value(Annotable, Immutable):
        a = is_int

    op = Value(1)
    with pytest.raises(AttributeError):
        op.a = 3


class BaseValue(Annotable):
    i = is_int
    j = attribute(is_int)


class Value2(BaseValue):
    @attribute
    def k(self):
        return 3


class Value3(BaseValue):
    k = attribute(is_int, default=3)


class Value4(BaseValue):
    k = attribute(Option(is_int), default=None)


def test_annotable_with_dict_slot():
    class Flexible(Annotable):
        __slots__ = ("__dict__",)

    v = Flexible()
    v.a = 1
    v.b = 2
    assert v.a == 1
    assert v.b == 2


def test_annotable_attribute():
    with pytest.raises(ValidationError, match="too many positional arguments"):
        BaseValue(1, 2)

    v = BaseValue(1)
    assert v.__slots__ == ("i", "j")
    assert v.i == 1
    assert not hasattr(v, "j")
    v.j = 2
    assert v.j == 2

    with pytest.raises(ValidationError):
        v.j = "foo"


def test_annotable_attribute_init():
    assert Value2.__slots__ == ("k",)
    v = Value2(1)

    assert v.i == 1
    assert not hasattr(v, "j")
    v.j = 2
    assert v.j == 2
    assert v.k == 3

    v = Value3(1)
    assert v.k == 3

    v = Value4(1)
    assert v.k is None


def test_annotable_mutability_and_serialization():
    v_ = BaseValue(1)
    v_.j = 2
    v = BaseValue(1)
    v.j = 2
    assert v_ == v
    assert v_.j == v.j == 2

    assert repr(v) == "BaseValue(i=1)"
    w = pickle.loads(pickle.dumps(v))
    assert w.i == 1
    assert w.j == 2
    assert v == w

    v.j = 4
    assert v_ != v
    w = pickle.loads(pickle.dumps(v))
    assert w == v
    assert repr(w) == "BaseValue(i=1)"


def test_initialized_attribute_basics():
    class Value(Annotable):
        a = is_int

        @attribute
        def double_a(self):
            return 2 * self.a

    op = Value(1)
    assert op.a == 1
    assert op.double_a == 2
    assert len(Value.__attributes__) == 1
    assert "double_a" in Value.__slots__


def test_initialized_attribute_with_validation():
    class Value(Annotable):
        a = is_int

        @attribute(int)
        def double_a(self):
            return 2 * self.a

    op = Value(1)
    assert op.a == 1
    assert op.double_a == 2
    assert len(Value.__attributes__) == 1
    assert "double_a" in Value.__slots__

    op.double_a = 3
    assert op.double_a == 3

    with pytest.raises(ValidationError):
        op.double_a = "foo"


def test_initialized_attribute_mixed_with_classvar():
    class Value(Annotable):
        arg = is_int

        shape = "like-arg"
        dtype = "like-arg"

    class Reduction(Value):
        shape = "scalar"

    class Variadic(Value):
        @attribute
        def shape(self):
            if self.arg > 10:
                return "columnar"
            else:
                return "scalar"

    r = Reduction(1)
    assert r.shape == "scalar"
    assert "shape" not in r.__slots__

    v = Variadic(1)
    assert v.shape == "scalar"
    assert "shape" in v.__slots__

    v = Variadic(100)
    assert v.shape == "columnar"
    assert "shape" in v.__slots__


def test_composition_of_annotable_and_singleton() -> None:
    class AnnSing(Annotable, Singleton):
        value = CoercedTo(int)

    class SingAnn(Singleton, Annotable):
        # this is the preferable method resolution order
        value = CoercedTo(int)

    # arguments looked up after validation
    obj1 = AnnSing("3")
    assert AnnSing("3") is obj1
    assert AnnSing(3) is obj1
    assert AnnSing(3.0) is obj1

    # arguments looked up before validation
    obj2 = SingAnn("3")
    assert SingAnn("3") is obj2
    obj3 = SingAnn(3)
    assert obj3 is not obj2
    assert SingAnn(3) is obj3


def test_concrete():
    assert BetweenWithCalculated.__mro__ == (
        BetweenWithCalculated,
        Concrete,
        Immutable,
        Comparable,
        Annotable,
        Abstract,
        object,
    )

    assert BetweenWithCalculated.__create__.__func__ is Annotable.__create__.__func__
    assert BetweenWithCalculated.__eq__ is Comparable.__eq__
    assert BetweenWithCalculated.__argnames__ == ("value", "lower", "upper")

    # annotable
    obj = BetweenWithCalculated(10, lower=5, upper=15)
    obj2 = BetweenWithCalculated(10, lower=5, upper=15)
    assert obj.value == 10
    assert obj.lower == 5
    assert obj.upper == 15
    assert obj.calculated == 15
    assert obj == obj2
    assert obj is not obj2
    assert obj != (10, 5, 15)
    assert obj.__args__ == (10, 5, 15)
    assert obj.args == (10, 5, 15)
    assert obj.argnames == ("value", "lower", "upper")

    # immutable
    with pytest.raises(AttributeError):
        obj.value = 11

    # hashable
    assert {obj: 1}.get(obj) == 1

    # weakrefable
    ref = weakref.ref(obj)
    assert ref() == obj

    # serializable
    assert pickle.loads(pickle.dumps(obj)) == obj


def test_composition_of_concrete_and_singleton():
    class ConcSing(Concrete, Singleton):
        value = CoercedTo(int)

    class SingConc(Singleton, Concrete):
        value = CoercedTo(int)

    # arguments looked up after validation
    obj = ConcSing("3")
    assert ConcSing("3") is obj
    assert ConcSing(3) is obj
    assert ConcSing(3.0) is obj

    # arguments looked up before validation
    obj = SingConc("3")
    assert SingConc("3") is obj
    obj2 = SingConc(3)
    assert obj2 is not obj
    assert SingConc(3) is obj2


def test_init_subclass_keyword_arguments():
    class Test(Annotable):
        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__()
            cls.kwargs = kwargs

    class Test2(Test, something="value", value="something"):
        pass

    assert Test2.kwargs == {"something": "value", "value": "something"}


def test_argument_order_using_optional_annotations():
    class Case1(Annotable):
        results: Optional[tuple[int]] = ()
        default: Optional[int] = None

    class SimpleCase1(Case1):
        base: int
        cases: Optional[tuple[int]] = ()

    class Case2(Annotable):
        results = optional(TupleOf(is_int), default=())
        default = optional(is_int)

    class SimpleCase2(Case1):
        base = is_int
        cases = optional(TupleOf(is_int), default=())

    assert (
        SimpleCase1.__argnames__
        == SimpleCase2.__argnames__
        == ("base", "cases", "results", "default")
    )


def test_annotable_with_optional_coercible_typehint():
    class Example(Annotable):
        value: Optional[MyInt] = None

    assert Example().value is None
    assert Example(None).value is None
    assert Example(1).value == 1
    assert isinstance(Example(1).value, MyInt)


def test_error_message(snapshot):
    class Example(Annotable):
        a: int
        b: int = 0
        c: str = "foo"
        d: Optional[float] = None
        e: tuple[int, ...] = (1, 2, 3)
        f: As[int] = 1

    with pytest.raises(ValidationError) as exc_info:
        Example("1", "2", "3", "4", "5", [])

    # assert "Failed" in str(exc_info.value)

    if sys.version_info >= (3, 11):
        target = "error_message_py311.txt"
    else:
        target = "error_message.txt"
    snapshot.assert_match(str(exc_info.value), target)



from __future__ import annotations

import numpy as np
import pandas as pd
import pandas.testing as tm
import pytest

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
from ibis.legacy.udf.vectorized import analytic, elementwise, reduction

dd = pytest.importorskip("dask.dataframe")


@pytest.fixture
def df(npartitions):
    return dd.from_pandas(
        pd.DataFrame(
            {
                "a": list("abc"),
                "b": [1, 2, 3],
                "c": [4.0, 5.0, 6.0],
                "key": list("aab"),
            }
        ),
        npartitions=npartitions,
    )


@pytest.fixture
def df2(npartitions):
    # df with some randomness
    return dd.from_pandas(
        pd.DataFrame(
            {
                "a": np.arange(4, dtype=float).tolist() + np.random.rand(3).tolist(),
                "b": np.arange(4, dtype=float).tolist() + np.random.rand(3).tolist(),
                "c": np.arange(7, dtype=int).tolist(),
                "d": list("aaaaddd"),
                "key": list("ddeefff"),
            }
        ),
        npartitions=npartitions,
    )


@pytest.fixture
def df_timestamp(npartitions):
    df = pd.DataFrame(
        {
            "a": list(range(10)),
            "b": list("wwwwwxxxxx"),
            "c": list("yyyzzzyyzz"),
        }
    )
    df["a"] = df.a.astype(pd.DatetimeTZDtype(tz="UTC"))
    return dd.from_pandas(
        df,
        npartitions=npartitions,
    )


@pytest.fixture
def con(df, df2, df_timestamp):
    return ibis.dask.connect({"df": df, "df2": df2, "df_timestamp": df_timestamp})


@pytest.fixture
def t(con):
    return con.table("df")


@pytest.fixture
def t2(con):
    return con.table("df2")


@pytest.fixture
def t_timestamp(con):
    return con.table("df_timestamp")


# -------------
# UDF Functions
# -------------


@elementwise(input_type=["string"], output_type="int64")
def my_string_length(series, **kwargs):
    return series.str.len() * 2


@elementwise(input_type=[dt.double, dt.double], output_type=dt.double)
def my_add(series1, series2, **kwargs):
    return series1 + series2


@reduction(["double"], "double")
def my_mean(series):
    return series.mean()


@reduction(
    input_type=[dt.Timestamp(timezone="UTC")],
    output_type=dt.Timestamp(timezone="UTC"),
)
def my_tz_min(series):
    return series.min()


@elementwise(
    input_type=[dt.Timestamp(timezone="UTC")],
    output_type=dt.Timestamp(timezone="UTC"),
)
def my_tz_add_one(series):
    return series + pd.Timedelta(1, unit="D")


@reduction(input_type=[dt.string], output_type=dt.int64)
def my_string_length_sum(series, **kwargs):
    return (series.str.len() * 2).sum()


@reduction(input_type=[dt.double, dt.double], output_type=dt.double)
def my_corr(lhs, rhs, **kwargs):
    return lhs.corr(rhs)


@elementwise([dt.double], dt.double)
def add_one(x):
    return x + 1.0


@elementwise([dt.double], dt.double)
def times_two(x):
    return x * 2.0


@analytic(input_type=["double"], output_type="double")
def zscore(series):
    return (series - series.mean()) / series.std()


@reduction(
    input_type=[dt.double],
    output_type=dt.Array(dt.double),
)
def collect(series):
    return list(series)


# -----
# Tests
# -----


def test_udf(t, df):
    expr = my_string_length(t.a)

    assert isinstance(expr, ir.Column)

    result = expr.execute()
    expected = df.a.str.len().mul(2).compute()

    tm.assert_series_equal(result, expected, check_names=False, check_index=False)


def test_multiple_argument_udf(t, df):
    expr = my_add(t.b, t.c)

    assert isinstance(expr, ir.Column)
    assert isinstance(expr, ir.NumericColumn)
    assert isinstance(expr, ir.FloatingColumn)

    result = expr.execute()
    expected = (df.b + df.c).compute()
    tm.assert_series_equal(result, expected, check_index=False, check_names=False)


def test_multiple_argument_udf_group_by(t):
    expr = t.group_by(t.key).aggregate(my_add=my_add(t.b, t.c).sum())

    assert isinstance(expr, ir.Table)
    assert isinstance(expr.my_add, ir.Column)
    assert isinstance(expr.my_add, ir.NumericColumn)
    assert isinstance(expr.my_add, ir.FloatingColumn)

    result = expr.execute()
    expected = pd.DataFrame(
        {"key": list("ab"), "my_add": [sum([1.0 + 4.0, 2.0 + 5.0]), 3.0 + 6.0]}
    )
    tm.assert_frame_equal(
        result.reset_index(drop=True), expected.reset_index(drop=True)
    )


def test_udaf(t):
    expr = my_string_length_sum(t.a)

    assert isinstance(expr, ir.Scalar)

    result = expr.execute()
    expected = t.a.execute().str.len().mul(2).sum()
    assert result == expected


def test_udaf_analytic_tzcol(t_timestamp, df_timestamp):
    expr = my_tz_min(t_timestamp.a)

    result = expr.execute()

    expected = my_tz_min.func(df_timestamp.a.compute())
    assert result == expected


def test_udaf_elementwise_tzcol(t_timestamp, df_timestamp):
    expr = my_tz_add_one(t_timestamp.a)

    result = expr.execute().reset_index(drop=True)

    expected = my_tz_add_one.func(df_timestamp.a.compute())
    tm.assert_series_equal(result, expected, check_index=False, check_names=False)


def test_udaf_analytic(t, df):
    expr = zscore(t.c)

    assert isinstance(expr, ir.Column)

    result = expr.execute()

    def f(s):
        return s.sub(s.mean()).div(s.std())

    expected = (f(df.c)).compute()
    tm.assert_series_equal(result, expected, check_index=False, check_names=False)


def test_udaf_analytic_group_by(t, df):
    expr = zscore(t.c).over(ibis.window(group_by=t.key))

    assert isinstance(expr, ir.Column)

    result = expr.execute()

    def f(s):
        return s.sub(s.mean()).div(s.std())

    expected = df.groupby("key").c.transform(f).compute()
    # We don't check names here because the udf is used "directly".
    # We could potentially special case this and set the name directly
    # if the udf is only being run on one column.
    tm.assert_series_equal(
        result.sort_index(), expected.sort_index(), check_names=False, check_index=False
    )


def test_udaf_group_by(t2, df2):
    expr = t2.group_by(t2.key).aggregate(my_corr=my_corr(t2.a, t2.b))

    result = expr.execute().sort_values("key").reset_index(drop=True)

    dfi = df2.set_index("key").compute()
    expected = pd.DataFrame(
        {
            "key": list("def"),
            "my_corr": [
                dfi.loc[value, "a"].corr(dfi.loc[value, "b"]) for value in "def"
            ],
        }
    )

    tm.assert_frame_equal(
        result.reset_index(drop=True), expected.reset_index(drop=True)
    )


def test_udaf_group_by_multikey(t2, df2):
    expr = t2.group_by([t2.key, t2.d]).aggregate(my_corr=my_corr(t2.a, t2.b))

    result = expr.execute().sort_values("key").reset_index(drop=True)

    dfi = df2.set_index("key").compute()
    expected = pd.DataFrame(
        {
            "key": list("def"),
            "d": list("aad"),
            "my_corr": [
                dfi.loc[value, "a"].corr(dfi.loc[value, "b"]) for value in "def"
            ],
        }
    )
    tm.assert_frame_equal(
        result.reset_index(drop=True), expected.reset_index(drop=True)
    )


def test_udaf_group_by_multikey_tzcol(t_timestamp, df_timestamp):
    expr = t_timestamp.group_by([t_timestamp.b, t_timestamp.c]).aggregate(
        my_min_time=my_tz_min(t_timestamp.a)
    )

    result = expr.execute().sort_values("b").reset_index(drop=True)
    expected = (
        df_timestamp.groupby(["b", "c"])
        .min()
        .reset_index()
        .rename(columns={"a": "my_min_time"})
        .compute()
    )
    tm.assert_frame_equal(
        result.reset_index(drop=True), expected.reset_index(drop=True)
    )


def test_compose_udfs(t2, df2):
    expr = times_two(add_one(t2.a))
    result = expr.execute().reset_index(drop=True)
    expected = df2.a.add(1.0).mul(2.0).compute()
    tm.assert_series_equal(result, expected, check_names=False, check_index=False)


def test_udaf_window(t2, df2):
    window = ibis.trailing_window(2, order_by="a", group_by="key")
    expr = t2.mutate(rolled=my_mean(t2.b).over(window))
    result = expr.execute().sort_values(["key", "a"])
    expected = (
        df2.compute()
        .sort_values(["key", "a"])
        .assign(
            rolled=lambda df: df.groupby("key")
            .b.rolling(3, min_periods=1)
            .mean()
            .reset_index(level=0, drop=True)
        )
    )
    tm.assert_frame_equal(result, expected)


def test_array_return_type_reduction(t, df):
    """Tests reduction UDF returning an array."""
    expr = collect(t.b)
    result = expr.execute()
    expected = df.b.compute().tolist()
    assert list(result) == expected


def test_array_return_type_reduction_window(t, df):
    """Tests reduction UDF returning an array, used over a window."""
    expr = collect(t.b).over(ibis.window())
    result = expr.execute()
    expected_raw = df.b.compute().tolist()
    expected = pd.Series([expected_raw] * len(df))
    tm.assert_series_equal(result, expected, check_index=False, check_names=False)


def test_array_return_type_reduction_group_by(t, df):
    """Tests reduction UDF returning an array, used in a grouped agg."""
    expr = t.group_by(t.key).aggregate(quantiles_col=collect(t.b))
    result = expr.execute()

    df = df.compute()  # Convert to Pandas
    expected_col = df.groupby(df.key).b.agg(lambda s: s.tolist())
    expected = pd.DataFrame({"quantiles_col": expected_col}).reset_index()

    tm.assert_frame_equal(
        result.sort_values("key").reset_index(drop=True),
        expected.sort_values("key").reset_index(drop=True),
    )


def test_elementwise_udf_with_many_args(t2):
    @elementwise(input_type=[dt.double] * 16 + [dt.int32] * 8, output_type=dt.double)
    def my_udf(
        c1,
        c2,
        c3,
        c4,
        c5,
        c6,
        c7,
        c8,
        c9,
        c10,
        c11,
        c12,
        c13,
        c14,
        c15,
        c16,
        c17,
        c18,
        c19,
        c20,
        c21,
        c22,
        c23,
        c24,
    ):
        return c1

    expr = my_udf(*([t2.a] * 8 + [t2.b] * 8 + [t2.c] * 8))
    result = expr.execute()
    expected = t2.a.execute()

    tm.assert_series_equal(result, expected, check_names=False, check_index=False)


# -----------------
# Test raied errors
# -----------------


def test_udaf_parameter_mismatch():
    with pytest.raises(TypeError):

        @reduction(input_type=[dt.double], output_type=dt.double)
        def my_corr(lhs, rhs, **kwargs):
            pass


def test_udf_parameter_mismatch():
    with pytest.raises(TypeError):

        @reduction(input_type=[], output_type=dt.double)
        def my_corr2(lhs, **kwargs):
            pass


def test_udf_error(t):
    @elementwise(input_type=[dt.double], output_type=dt.double)
    def error_udf(s):
        raise ValueError("xxx")

    with pytest.raises(ValueError):
        error_udf(t.c).execute()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from ibis.expr import types as ir

from public import public

import ibis.expr.operations as ops
from ibis.expr.types.generic import Column, Scalar, Value


@public
class BinaryValue(Value):
    def hashbytes(
        self,
        how: Literal["md5", "sha1", "sha256", "sha512"] = "sha256",
    ) -> ir.BinaryValue:
        """Compute the binary hash value of `arg`.

        Parameters
        ----------
        how
            Hash algorithm to use

        Returns
        -------
        BinaryValue
            Binary expression
        """
        return ops.HashBytes(self, how).to_expr()


@public
class BinaryScalar(Scalar, BinaryValue):
    pass


@public
class BinaryColumn(Column, BinaryValue):
    pass



from __future__ import annotations

import pytest

import ibis
from ibis.backends.impala import ddl
from ibis.backends.impala.ddl import (
    CTAS,
    CreateTableWithSchema,
    DropTable,
    InsertSelect,
)


@pytest.fixture
def t(mockcon):
    return mockcon.table("functional_alltypes")


def test_drop_table_compile(snapshot):
    statement = DropTable("foo", database="bar", must_exist=True)
    query = statement.compile()
    snapshot.assert_match(query, "out1.sql")

    statement = DropTable("foo", database="bar", must_exist=False)
    query = statement.compile()
    snapshot.assert_match(query, "out2.sql")


def test_select_basics(t, snapshot):
    name = "testing123456"

    expr = t.limit(10)

    select = str(ibis.to_sql(expr, "impala"))
    stmt = InsertSelect(name, select, database="foo")
    result = stmt.compile()
    snapshot.assert_match(result, "out1.sql")

    stmt = InsertSelect(name, select, database="foo", overwrite=True)
    result = stmt.compile()
    snapshot.assert_match(result, "out2.sql")


def test_load_data_unpartitioned(snapshot):
    path = "/path/to/data"
    stmt = ddl.LoadData("functional_alltypes", path, database="foo")

    result = stmt.compile()
    snapshot.assert_match(result, "out1.sql")

    stmt.overwrite = True
    result = stmt.compile()
    snapshot.assert_match(result, "out2.sql")


def test_load_data_partitioned(snapshot):
    path = "/path/to/data"
    part = {"year": 2007, "month": 7}
    part_schema = ibis.schema([("year", "int32"), ("month", "int32")])
    stmt = ddl.LoadData(
        "functional_alltypes",
        path,
        database="foo",
        partition=part,
        partition_schema=part_schema,
    )

    result = stmt.compile()
    snapshot.assert_match(result, "out1.sql")

    stmt.overwrite = True
    result = stmt.compile()
    snapshot.assert_match(result, "out2.sql")


def test_cache_table_pool_name(snapshot):
    statement = ddl.CacheTable("foo", database="bar")
    query = statement.compile()
    snapshot.assert_match(query, "out1.sql")

    statement = ddl.CacheTable("foo", database="bar", pool="my_pool")
    query = statement.compile()
    snapshot.assert_match(query, "out2.sql")


@pytest.fixture
def part_schema():
    return ibis.schema([("year", "int32"), ("month", "int32")])


@pytest.fixture
def table_name():
    return "tbl"


def test_add_partition(part_schema, table_name, snapshot):
    stmt = ddl.AddPartition(table_name, {"year": 2007, "month": 4}, part_schema)

    result = stmt.compile()
    snapshot.assert_match(result, "out.sql")


def test_add_partition_string_key(snapshot):
    part_schema = ibis.schema([("foo", "int32"), ("bar", "string")])
    stmt = ddl.AddPartition("tbl", {"foo": 5, "bar": "qux"}, part_schema)

    result = stmt.compile()
    snapshot.assert_match(result, "out.sql")


def test_drop_partition(part_schema, table_name, snapshot):
    stmt = ddl.DropPartition(table_name, {"year": 2007, "month": 4}, part_schema)

    result = stmt.compile()
    snapshot.assert_match(result, "out.sql")


def test_add_partition_with_props(part_schema, table_name, snapshot):
    props = {"location": "/users/foo/my-data"}
    stmt = ddl.AddPartition(
        table_name, {"year": 2007, "month": 4}, part_schema, **props
    )

    result = stmt.compile()
    snapshot.assert_match(result, "out.sql")


def test_alter_partition_properties(part_schema, table_name, snapshot):
    part = {"year": 2007, "month": 4}

    def _get_ddl_string(props):
        stmt = ddl.AlterPartition(table_name, part, part_schema, **props)
        return stmt.compile()

    result = _get_ddl_string({"location": "/users/foo/my-data"})
    snapshot.assert_match(result, "out1.sql")

    result = _get_ddl_string({"format": "avro"})
    snapshot.assert_match(result, "out2.sql")

    result = _get_ddl_string({"tbl_properties": {"bar": 2, "foo": "1"}})
    snapshot.assert_match(result, "out3.sql")

    result = _get_ddl_string({"serde_properties": {"baz": 3}})
    snapshot.assert_match(result, "out4.sql")


def test_alter_table_properties(part_schema, table_name, snapshot):
    part = {"year": 2007, "month": 4}

    def _get_ddl_string(props):
        stmt = ddl.AlterPartition(table_name, part, part_schema, **props)
        return stmt.compile()

    result = _get_ddl_string({"location": "/users/foo/my-data"})
    snapshot.assert_match(result, "out1.sql")

    result = _get_ddl_string({"format": "avro"})
    snapshot.assert_match(result, "out2.sql")

    result = _get_ddl_string({"tbl_properties": {"bar": 2, "foo": "1"}})
    snapshot.assert_match(result, "out3.sql")

    result = _get_ddl_string({"serde_properties": {"baz": 3}})
    snapshot.assert_match(result, "out4.sql")


@pytest.fixture
def expr(t):
    return t[t.bigint_col > 0]


def test_create_external_table_as(mockcon, snapshot):
    path = "/path/to/table"
    select = ibis.to_sql(mockcon.table("test1"), "impala")
    statement = CTAS(
        "another_table",
        select,
        external=True,
        can_exist=False,
        path=path,
        database="foo",
    )
    result = statement.compile()
    snapshot.assert_match(result, "out.sql")


def test_create_table_with_location_compile(snapshot):
    path = "/path/to/table"
    schema = ibis.schema([("foo", "string"), ("bar", "int8"), ("baz", "int16")])
    statement = CreateTableWithSchema(
        "another_table",
        schema,
        can_exist=False,
        format="parquet",
        path=path,
        database="foo",
    )
    result = statement.compile()
    snapshot.assert_match(result, "out.sql")


def test_create_table_like_parquet(snapshot):
    directory = "/path/to/"
    path = "/path/to/parquetfile"
    statement = ddl.CreateTableParquet(
        "new_table",
        directory,
        example_file=path,
        can_exist=True,
        database="foo",
    )

    result = statement.compile()
    snapshot.assert_match(result, "out.sql")


def test_create_table_parquet_like_other(snapshot):
    # alternative to "LIKE PARQUET"
    directory = "/path/to/"
    example_table = "db.other"

    statement = ddl.CreateTableParquet(
        "new_table",
        directory,
        example_table=example_table,
        can_exist=True,
        database="foo",
    )

    result = statement.compile()
    snapshot.assert_match(result, "out.sql")


def test_create_table_parquet_with_schema(snapshot):
    directory = "/path/to/"

    schema = ibis.schema([("foo", "string"), ("bar", "int8"), ("baz", "int16")])

    statement = ddl.CreateTableParquet(
        "new_table",
        directory,
        schema=schema,
        external=True,
        can_exist=True,
        database="foo",
    )

    result = statement.compile()
    snapshot.assert_match(result, "out.sql")


def test_create_table_delimited(snapshot):
    path = "/path/to/files/"
    schema = ibis.schema(
        [
            ("a", "string"),
            ("b", "int32"),
            ("c", "double"),
            ("d", "decimal(12, 2)"),
        ]
    )

    stmt = ddl.CreateTableDelimited(
        "new_table",
        path,
        schema,
        delimiter="|",
        escapechar="\\",
        lineterminator="\0",
        database="foo",
        can_exist=True,
    )

    result = stmt.compile()
    snapshot.assert_match(result, "out.sql")


def test_create_external_table_avro(snapshot):
    path = "/path/to/files/"

    avro_schema = {
        "fields": [
            {"name": "a", "type": "string"},
            {"name": "b", "type": "int"},
            {"name": "c", "type": "double"},
            {
                "type": "bytes",
                "logicalType": "decimal",
                "precision": 4,
                "scale": 2,
                "name": "d",
            },
        ],
        "name": "my_record",
        "type": "record",
    }

    stmt = ddl.CreateTableAvro(
        "new_table", path, avro_schema, database="foo", can_exist=True
    )

    result = stmt.compile()
    snapshot.assert_match(result, "out.sql")


def test_create_table_parquet(expr, snapshot):
    statement = _create_table("some_table", expr, database="bar", can_exist=False)
    result = statement.compile()
    snapshot.assert_match(result, "out.sql")


def test_no_overwrite(expr, snapshot):
    statement = _create_table("tname", expr, can_exist=True)
    result = statement.compile()
    snapshot.assert_match(result, "out.sql")


def test_avro_other_formats(t, snapshot):
    statement = _create_table("tname", t, format="avro", can_exist=True)
    result = statement.compile()
    snapshot.assert_match(result, "out.sql")

    with pytest.raises(ValueError):
        _create_table("tname", t, format="foo")


def _create_table(table_name, expr, database=None, can_exist=False, format="parquet"):
    select = str(ibis.to_sql(expr, dialect="impala"))
    return CTAS(
        table_name, select, database=database, format=format, can_exist=can_exist
    )



from __future__ import annotations

import ibis

from .conftest import add_date, tpch_test


@tpch_test
def test_tpc_h06(lineitem):
    "Forecasting Revenue Change Query (Q6)"
    DATE = "1994-01-01"
    DISCOUNT = 0.06
    QUANTITY = 24

    q = lineitem
    discount_min = round(DISCOUNT - 0.01, 2)
    discount_max = round(DISCOUNT + 0.01, 2)
    q = q.filter(
        [
            q.l_shipdate >= ibis.date(DATE),
            q.l_shipdate < add_date(DATE, dy=1),
            q.l_discount.between(discount_min, discount_max),
            q.l_quantity < QUANTITY,
        ]
    )
    q = q.aggregate(revenue=(q.l_extendedprice * q.l_discount).sum())
    return q



from __future__ import annotations

from collections import OrderedDict

import pytest

import ibis
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import _
from ibis.expr.tests.test_newrels import join_tables
from ibis.tests.util import assert_pickle_roundtrip


@pytest.fixture
def t():
    return ibis.table(dict(a="struct<b: float, c: string>", d="string"), name="t")


@pytest.fixture
def s():
    return ibis.table(dict(a="struct<f: float, g: string>"), name="s")


def test_struct_operations():
    value = OrderedDict(
        [
            ("a", 1),
            ("b", list("abc")),
            ("c", OrderedDict([("foo", [1.0, 2.0])])),
        ]
    )
    expr = ibis.literal(value)
    assert isinstance(expr, ir.StructValue)
    assert isinstance(expr["b"], ir.ArrayValue)
    assert isinstance(expr["a"].op(), ops.StructField)


def test_struct_getattr():
    expr = ibis.struct({"a": 1, "b": 2})
    assert isinstance(expr.a, ir.IntegerValue)
    assert expr.a.get_name() == "a"
    with pytest.raises(AttributeError, match="bad"):
        expr.bad  # # noqa: B018


def test_struct_tab_completion():
    t = ibis.table([("struct_col", "struct<my_field: string, for: int64>")])
    # Only valid python identifiers in getattr completions
    attrs = dir(t.struct_col)
    assert "my_field" in attrs
    assert "for" not in attrs
    # All fields in getitem completions
    items = t.struct_col._ipython_key_completions_()
    assert {"my_field", "for"}.issubset(items)


def test_struct_pickle():
    struct_scalar_expr = ibis.literal(OrderedDict([("fruit", "pear"), ("weight", 0)]))

    assert_pickle_roundtrip(struct_scalar_expr)


def test_lift(t):
    assert t.a.lift().equals(t[_.a.b, _.a.c])


def test_unpack_from_table(t):
    assert t.unpack("a").equals(t[_.a.b, _.a.c, _.d])


def test_lift_join(t, s):
    join = t.join(s, t.d == s.a.g)
    result = join.a_right.lift()

    with join_tables(t, s) as (r1, r2):
        join = ops.JoinChain(
            first=r1,
            rest=[
                ops.JoinLink("inner", r2, [r1.d == r2.a.g]),
            ],
            values={"f": r2.a.f, "g": r2.a.g},
        )
        assert result.op() == join


def test_unpack_join_from_table(t, s):
    join = t.join(s, t.d == s.a.g)
    result = join.unpack("a_right")
    expected = join[_.a, _.d, _.a_right.f, _.a_right.g]
    assert result.equals(expected)


def test_nested_lift():
    t = ibis.table(
        {"a": "struct<b:struct<x: int, y: int>, c: string>", "d": "string"},
        name="t",
    )
    expr = t.a.b.lift()
    assert expr.schema() == ibis.schema({"x": "int", "y": "int"})



from __future__ import annotations

import os
import shutil
from posixpath import join as pjoin

import pytest

import ibis
from ibis import util
from ibis.tests.util import assert_equal

pyspark = pytest.importorskip("pyspark")


@pytest.fixture
def temp_view(con) -> str:
    name = util.gen_name("view")
    yield name
    con.drop_view(name, force=True)


def test_create_exists_view(con, alltypes, temp_view):
    assert temp_view not in con.list_tables()

    t1 = alltypes.group_by("string_col").size()
    t2 = con.create_view(temp_view, t1)

    assert temp_view in con.list_tables()
    # just check it works for now
    assert t2.execute() is not None


def test_drop_non_empty_database(con, alltypes, temp_table_db):
    temp_database, temp_table = temp_table_db
    con.create_table(temp_table, alltypes, database=temp_database)
    assert temp_table in con.list_tables(database=temp_database)

    with pytest.raises(pyspark.sql.utils.AnalysisException):
        con.drop_database(temp_database)


@pytest.fixture
def temp_base():
    base = pjoin(f"/tmp/{util.gen_name('pyspark_testing')}", util.gen_name("temp_base"))
    yield base
    shutil.rmtree(base, ignore_errors=True)


@pytest.fixture
def temp_db(con, temp_base):
    name = util.gen_name("temp_db")
    yield pjoin(temp_base, name)
    con.drop_database(name, force=True)


def test_create_database_with_location(con, temp_db):
    base = os.path.dirname(temp_db)
    name = os.path.basename(temp_db)
    con.create_database(name, path=temp_db)
    assert os.path.exists(base)


def test_drop_table_not_exist(con):
    non_existent_table = f"ibis_table_{util.guid()}"
    with pytest.raises(pyspark.sql.utils.AnalysisException):
        con.drop_table(non_existent_table)
    con.drop_table(non_existent_table, force=True)


def test_truncate_table(con, alltypes, temp_table):
    expr = alltypes.limit(1)

    con.create_table(temp_table, obj=expr)
    con.truncate_table(temp_table)

    t = con.table(temp_table)
    nrows = t.count().execute()
    assert not nrows


def test_ctas_from_table_expr(con, alltypes, temp_table_db):
    expr = alltypes
    db, table_name = temp_table_db

    con.create_table(table_name, expr, database=db)


def test_create_empty_table(con, temp_table):
    schema = ibis.schema(
        [
            ("a", "string"),
            ("b", "timestamp"),
            ("c", "decimal(12, 8)"),
            ("d", "double"),
        ]
    )

    con.create_table(temp_table, schema=schema)

    result_schema = con.get_schema(temp_table)
    assert_equal(result_schema, schema)

    assert con.table(temp_table).execute().empty


def test_insert_table(con, alltypes, temp_table, test_data_db):
    expr = alltypes
    db = test_data_db

    con.create_table(temp_table, expr.limit(0), database=db)
    con.insert(temp_table, expr.limit(10), database=db)
    assert con.table(temp_table).count().execute() == 10

    # Overwrite and verify only 10 rows now
    con.insert(temp_table, expr.limit(10), overwrite=True)
    assert con.table(temp_table).count().execute() == 10

    con.insert(temp_table, expr.limit(10), database=db, overwrite=False)
    assert con.table(temp_table).count().execute() == 20


def test_insert_validate_types(con, alltypes, test_data_db, temp_table):
    db = test_data_db

    expr = alltypes
    con.create_table(
        temp_table,
        schema=expr["tinyint_col", "int_col", "string_col"].schema(),
        database=db,
    )

    to_insert = expr[
        expr.tinyint_col, expr.smallint_col.name("int_col"), expr.string_col
    ]
    con.insert(temp_table, to_insert.limit(10))

    to_insert = expr[
        expr.tinyint_col,
        expr.smallint_col.cast("int32").name("int_col"),
        expr.string_col,
    ]
    con.insert(temp_table, to_insert.limit(10))


def test_compute_stats(con, alltypes, temp_table):
    con.create_table(temp_table, alltypes)
    con.compute_stats(temp_table)
    con.compute_stats(temp_table, noscan=True)


@pytest.fixture
def created_view(con, alltypes):
    name = util.guid()
    expr = alltypes.limit(10)
    con.create_view(name, expr)
    return name


def test_drop_view(con, created_view):
    con.drop_view(created_view)
    assert created_view not in con.list_tables()


@pytest.fixture
def table(con, temp_database):
    table_name = f"table_{util.guid()}"
    schema = ibis.schema([("foo", "string"), ("bar", "int64")])
    yield con.create_table(
        table_name, database=temp_database, schema=schema, format="parquet"
    )
    con.drop_table(table_name, database=temp_database)


@pytest.fixture
def keyword_t(con):
    yield "distinct"
    con.drop_table("distinct")


def test_create_table_reserved_identifier(con, alltypes, keyword_t):
    expr = alltypes
    expected = expr.count().execute()
    t = con.create_table(keyword_t, expr)
    result = t.count().execute()
    assert result == expected



from __future__ import annotations

import collections
import itertools
import math
from collections.abc import Callable, Hashable, Iterable, Iterator, Mapping
from typing import Any, TypeVar

from ibis.common.graph import Node
from ibis.util import promote_list

K = TypeVar("K", bound=Hashable)


class DisjointSet(Mapping[K, set[K]]):
    """Disjoint set data structure.

    Also known as union-find data structure. It is a data structure that keeps
    track of a set of elements partitioned into a number of disjoint (non-overlapping)
    subsets. It provides near-constant-time operations to add new sets, to merge
    existing sets, and to determine whether elements are in the same set.

    Parameters
    ----------
    data :
        Initial data to add to the disjoint set.

    Examples
    --------
    >>> ds = DisjointSet()
    >>> ds.add(1)
    1
    >>> ds.add(2)
    2
    >>> ds.add(3)
    3
    >>> ds.union(1, 2)
    True
    >>> ds.union(2, 3)
    True
    >>> ds.find(1)
    1
    >>> ds.find(2)
    1
    >>> ds.find(3)
    1
    >>> ds.union(1, 3)
    False

    """

    __slots__ = ("_parents", "_classes")
    _parents: dict
    _classes: dict

    def __init__(self, data: Iterable[K] | None = None):
        self._parents = {}
        self._classes = {}
        if data is not None:
            for id in data:
                self.add(id)

    def __contains__(self, id) -> bool:
        """Check if the given id is in the disjoint set.

        Parameters
        ----------
        id :
            The id to check.

        Returns
        -------
        ined:
            True if the id is in the disjoint set, False otherwise.

        """
        return id in self._parents

    def __getitem__(self, id) -> set[K]:
        """Get the set of ids that are in the same class as the given id.

        Parameters
        ----------
        id :
            The id to get the class for.

        Returns
        -------
        class:
            The set of ids that are in the same class as the given id, including
            the given id.

        """
        id = self._parents[id]
        return self._classes[id]

    def __iter__(self) -> Iterator[K]:
        """Iterate over the ids in the disjoint set."""
        return iter(self._parents)

    def __len__(self) -> int:
        """Get the number of ids in the disjoint set."""
        return len(self._parents)

    def __eq__(self, other: object) -> bool:
        """Check if the disjoint set is equal to another disjoint set.

        Parameters
        ----------
        other :
            The other disjoint set to compare to.

        Returns
        -------
        equal:
            True if the disjoint sets are equal, False otherwise.

        """
        if not isinstance(other, DisjointSet):
            return NotImplemented
        return self._parents == other._parents

    def copy(self) -> DisjointSet:
        """Make a copy of the disjoint set.

        Returns
        -------
        copy:
            A copy of the disjoint set.

        """
        ds = DisjointSet()
        ds._parents = self._parents.copy()
        ds._classes = self._classes.copy()
        return ds

    def add(self, id: K) -> K:
        """Add a new id to the disjoint set.

        If the id is not in the disjoint set, it will be added to the disjoint set
        along with a new class containing only the given id.

        Parameters
        ----------
        id :
            The id to add to the disjoint set.

        Returns
        -------
        id:
            The id that was added to the disjoint set.

        """
        if id in self._parents:
            return self._parents[id]
        self._parents[id] = id
        self._classes[id] = {id}
        return id

    def find(self, id: K) -> K:
        """Find the root of the class that the given id is in.

        Also called as the canonicalized id or the representative id.

        Parameters
        ----------
        id :
            The id to find the canonicalized id for.

        Returns
        -------
        id:
            The canonicalized id for the given id.

        """
        return self._parents[id]

    def union(self, id1, id2) -> bool:
        """Merge the classes that the given ids are in.

        If the ids are already in the same class, this will return False. Otherwise
        it will merge the classes and return True.

        Parameters
        ----------
        id1 :
            The first id to merge the classes for.
        id2 :
            The second id to merge the classes for.

        Returns
        -------
        merged:
            True if the classes were merged, False otherwise.

        """
        # Find the root of each class
        id1 = self._parents[id1]
        id2 = self._parents[id2]
        if id1 == id2:
            return False

        # Merge the smaller eclass into the larger one, aka. union-find by size
        class1 = self._classes[id1]
        class2 = self._classes[id2]
        if len(class1) >= len(class2):
            id1, id2 = id2, id1
            class1, class2 = class2, class1

        # Update the parent pointers, this is called path compression but done
        # during the union operation to keep the find operation minimal
        for id in class1:
            self._parents[id] = id2

        # Do the actual merging and clear the other eclass
        class2 |= class1
        class1.clear()

        return True

    def connected(self, id1, id2):
        """Check if the given ids are in the same class.

        True if both ids have the same canonicalized id, False otherwise.

        Parameters
        ----------
        id1 :
            The first id to check.
        id2 :
            The second id to check.

        Returns
        -------
        connected:
            True if the ids are connected, False otherwise.

        """
        return self._parents[id1] == self._parents[id2]

    def verify(self):
        """Verify that the disjoint set is not corrupted.

        Check that each id's canonicalized id's class. In general corruption
        should not happen if the public API is used, but this is a sanity check
        to make sure that the internal data structures are not corrupted.

        Returns
        -------
        verified:
            True if the disjoint set is not corrupted, False otherwise.

        """
        for id in self._parents:
            if id not in self._classes[self._parents[id]]:
                raise RuntimeError(
                    f"DisjointSet is corrupted: {id} is not in its class"
                )


class Slotted:
    """A lightweight alternative to `ibis.common.grounds.Concrete`.

    This class is used to create immutable dataclasses with slots and a precomputed
    hash value for quicker dictionary lookups.
    """

    __slots__ = ("__precomputed_hash__",)
    __precomputed_hash__: int

    def __init__(self, *args):
        for name, value in itertools.zip_longest(self.__slots__, args):
            object.__setattr__(self, name, value)
        object.__setattr__(self, "__precomputed_hash__", hash(args))

    def __eq__(self, other):
        if self is other:
            return True
        if type(self) is not type(other):
            return NotImplemented
        for name in self.__slots__:
            if getattr(self, name) != getattr(other, name):
                return False
        return True

    def __hash__(self):
        return self.__precomputed_hash__

    def __setattr__(self, name, value):
        raise AttributeError("Can't set attributes on immutable ENode instance")


class Variable(Slotted):
    """A named capture in a pattern.

    Parameters
    ----------
    name : str
        The name of the variable.

    """

    __slots__ = ("name",)
    name: str

    def __init__(self, name: str):
        if name is None:
            raise ValueError("Variable name cannot be None")
        super().__init__(name)

    def __repr__(self):
        return f"${self.name}"

    def substitute(self, egraph, enode, subst):
        """Substitute the variable with the corresponding value in the substitution.

        Parameters
        ----------
        egraph : EGraph
            The egraph instance.
        enode : ENode
            The matched enode.
        subst : dict
            The substitution dictionary.

        Returns
        -------
        value : Any
            The substituted value.

        """
        return subst[self.name]


# Pattern corresponds to a selection which is flattened to a join of selections
class Pattern(Slotted):
    """A non-ground term, tree of enodes possibly containing variables.

    This class is used to represent a pattern in a query. The pattern is almost
    identical to an ENode, except that it can contain variables.

    Parameters
    ----------
    head : type
        The head or python type of the ENode to match against.
    args : tuple
        The arguments of the pattern. The arguments can be enodes, patterns,
        variables or leaf values.
    name : str, optional
        The name of the pattern which is used to refer to it in a rewrite rule.

    """

    __slots__ = ("head", "args", "name")
    head: type
    args: tuple
    name: str | None

    # TODO(kszucs): consider to raise if the pattern matches none
    def __init__(self, head, args, name=None, conditions=None):
        # TODO(kszucs): ensure that args are either patterns, variables or leaf values
        assert all(not isinstance(arg, (ENode, Node)) for arg in args)
        super().__init__(head, tuple(args), name)

    def matches_none(self):
        """Evaluate whether the pattern is guaranteed to match nothing.

        This can be evaluated before the matching loop starts, so eventually can
        be eliminated from the flattened query.
        """
        return len(self.head.__argnames__) != len(self.args)

    def matches_all(self):
        """Evaluate whether the pattern is guaranteed to match everything.

        This can be evaluated before the matching loop starts, so eventually can
        be eliminated from the flattened query.
        """
        return not self.matches_none() and all(
            isinstance(arg, Variable) for arg in self.args
        )

    def __repr__(self):
        argstring = ", ".join(map(repr, self.args))
        return f"P{self.head.__name__}({argstring})"

    def __rshift__(self, rhs):
        """Syntax sugar to create a rewrite rule."""
        return Rewrite(self, rhs)

    def __rmatmul__(self, name):
        """Syntax sugar to create a named pattern."""
        return self.__class__(self.head, self.args, name)

    def flatten(self, var=None, counter=None):
        """Recursively flatten the pattern to a join of selections.

        `Pattern(Add, (Pattern(Mul, ($x, 1)), $y))` is turned into a join of
        selections by introducing auxiliary variables where each selection gets
        executed as a dictionary lookup.

        In SQL terms this is equivalent to the following query:
        SELECT m.0 AS $x, a.1 AS $y FROM Add a JOIN Mul m ON a.0 = m.id WHERE m.1 = 1

        Parameters
        ----------
        var : Variable
            The variable to assign to the flattened pattern.
        counter : Iterator[int]
            The counter to generate unique variable names for auxiliary variables
            connecting the selections.

        Yields
        ------
        (var, pattern) : tuple[Variable, Pattern]
            The variable and the flattened pattern where the flattened pattern
            cannot contain any patterns just variables.

        """
        # TODO(kszucs): convert a pattern to a query object instead by flattening it
        counter = counter or itertools.count()

        if var is None:
            if self.name is None:
                var = Variable(next(counter))
            else:
                var = Variable(self.name)

        args = []
        for arg in self.args:
            if isinstance(arg, Pattern):
                if arg.name is None:
                    aux = Variable(next(counter))
                else:
                    aux = Variable(arg.name)
                yield from arg.flatten(aux, counter)
                args.append(aux)
            else:
                args.append(arg)

        yield (var, Pattern(self.head, args))

    def substitute(self, egraph, enode, subst):
        """Substitute the variables in the pattern with the corresponding values.

        Parameters
        ----------
        egraph : EGraph
            The egraph instance.
        enode : ENode
            The matched enode.
        subst : dict
            The substitution dictionary.

        Returns
        -------
        enode : ENode
            The substituted pattern which is a ground term aka. an ENode.

        """
        args = []
        for arg in self.args:
            if isinstance(arg, (Variable, Pattern)):
                arg = arg.substitute(egraph, enode, subst)
            args.append(arg)
        return ENode(self.head, tuple(args))


class DynamicApplier(Slotted):
    """A dynamic applier which calls a function to compute the result."""

    __slots__ = ("func",)
    func: Callable

    def substitute(self, egraph, enode, subst):
        kwargs = {k: v for k, v in subst.items() if isinstance(k, str)}
        result = self.func(egraph, enode, **kwargs)
        if not isinstance(result, ENode):
            raise TypeError(f"applier must return an ENode, got {type(result)}")
        return result


class Rewrite(Slotted):
    """A rewrite rule which matches a pattern and applies a pattern or a function."""

    __slots__ = ("matcher", "applier")
    matcher: Pattern
    applier: Callable | Pattern | Variable

    def __init__(self, matcher, applier):
        if callable(applier):
            applier = DynamicApplier(applier)
        elif not isinstance(applier, (Pattern, Variable)):
            raise TypeError(
                "applier must be a Pattern or a Variable returning an ENode"
            )
        super().__init__(matcher, applier)

    def __repr__(self):
        return f"{self.lhs} >> {self.rhs}"


class ENode(Slotted, Node):
    """A ground term which is a node in the EGraph, called ENode.

    Parameters
    ----------
    head : type
        The type of the Node the ENode represents.
    args : tuple
        The arguments of the ENode which are either ENodes or leaf values.

    """

    __slots__ = ("head", "args")
    head: type
    args: tuple

    def __init__(self, head, args):
        # TODO(kszucs): ensure that it is a ground term, this check should be removed
        assert all(not isinstance(arg, (Pattern, Variable)) for arg in args)
        super().__init__(head, tuple(args))

    @property
    def __argnames__(self):
        """Implementation for the `ibis.common.graph.Node` protocol."""
        return self.head.__argnames__

    @property
    def __args__(self):
        """Implementation for the `ibis.common.graph.Node` protocol."""
        return self.args

    def __repr__(self):
        argstring = ", ".join(map(repr, self.args))
        return f"E{self.head.__name__}({argstring})"

    def __lt__(self, other):
        return False

    @classmethod
    def from_node(cls, node: Any):
        """Convert an `ibis.common.graph.Node` to an `ENode`."""

        def mapper(node, _, **kwargs):
            return cls(node.__class__, kwargs.values())

        return node.map(mapper)[node]

    def to_node(self):
        """Convert the ENode back to an `ibis.common.graph.Node`."""

        def mapper(node, _, **kwargs):
            return node.head(**kwargs)

        return self.map(mapper)[self]


# TODO: move every E* into the Egraph so its API only uses Nodes
# TODO: track whether the egraph is saturated or not
# TODO: support parent classes in etables (Join <= InnerJoin)


class EGraph:
    __slots__ = ("_nodes", "_etables", "_eclasses")
    _nodes: dict
    _etables: collections.defaultdict
    _eclasses: DisjointSet

    def __init__(self):
        # store the nodes before converting them to enodes, so we can spare the initial
        # node traversal and omit the creation of enodes
        self._nodes = {}
        # map enode heads to their eclass ids and their arguments, this is required for
        # the relational e-matching (Node => dict[type, tuple[Union[ENode, Any], ...]])
        self._etables = collections.defaultdict(dict)
        # map enodes to their eclass, this is the heart of the egraph
        self._eclasses = DisjointSet()

    def __repr__(self):
        return f"EGraph({self._eclasses})"

    def _as_enode(self, node: Node) -> ENode:
        """Convert a node to an enode."""
        # order is important here since ENode is a subclass of Node
        if isinstance(node, ENode):
            return node
        elif isinstance(node, Node):
            return self._nodes.get(node) or ENode.from_node(node)
        else:
            raise TypeError(node)

    def add(self, node: Node) -> ENode:
        """Add a node to the egraph.

        The node is converted to an enode and added to the egraph. If the enode is
        already present in the egraph, then the canonical enode is returned.

        Parameters
        ----------
        node :
            The node to add to the egraph.

        Returns
        -------
        enode :
            The canonical enode.

        """
        enode = self._as_enode(node)
        if enode in self._eclasses:
            return self._eclasses.find(enode)

        args = []
        for arg in enode.args:
            if isinstance(arg, ENode):
                args.append(self.add(arg))
            else:
                args.append(arg)

        enode = ENode(enode.head, args)
        self._eclasses.add(enode)
        self._etables[enode.head][enode] = tuple(args)

        return enode

    def union(self, node1: Node, node2: Node) -> ENode:
        """Union two nodes in the egraph.

        The nodes are converted to enodes which must be present in the egraph.
        The eclasses of the nodes are merged and the canonical enode is returned.

        Parameters
        ----------
        node1 :
            The first node to union.
        node2 :
            The second node to union.

        Returns
        -------
        enode :
            The canonical enode.

        """
        enode1 = self._as_enode(node1)
        enode2 = self._as_enode(node2)
        return self._eclasses.union(enode1, enode2)

    def _match_args(self, args, patargs):
        """Match the arguments of an enode against a pattern's arguments.

        An enode matches a pattern if each of the arguments are:
        - both leaf values and equal
        - both enodes and in the same eclass
        - an enode and a variable, in which case the variable gets bound to the enode

        Parameters
        ----------
        args : tuple
            The arguments of the enode. Since an enode is a ground term, the arguments
            are either enodes or leaf values.
        patargs : tuple
            The arguments of the pattern. Since a pattern is a flat term (flattened
            using auxiliary variables), the arguments are either variables or leaf
            values.

        Returns
        -------
        dict[str, Any] :
            The mapping of variable names to enodes or leaf values.

        """
        subst = {}
        for arg, patarg in zip(args, patargs):
            if isinstance(patarg, Variable):
                if isinstance(arg, ENode):
                    subst[patarg.name] = self._eclasses.find(arg)
                else:
                    subst[patarg.name] = arg
            # TODO(kszucs): this is not needed since patarg is either a variable or a
            # leaf value due to the pattern flattening, though we may choose to
            # support this in the future
            # elif isinstance(arg, ENode):
            #     if self._eclasses.find(arg) != self._eclasses.find(arg):
            #         return None
            elif patarg != arg:
                return None
        return subst

    def match(self, pattern: Pattern) -> dict[ENode, dict[str, Any]]:
        """Match a pattern in the egraph.

        The pattern is converted to a conjunctive query (list of flat patterns) and
        matched against the relations represented by the egraph. This is called the
        relational e-matching.

        Parameters
        ----------
        pattern :
            The pattern to match in the egraph.

        Returns
        -------
        matches :
            A dictionary mapping the matched enodes to their substitutions.

        """
        # patterns could be reordered to match on the most selective one first
        patterns = dict(reversed(list(pattern.flatten())))
        if any(pat.matches_none() for pat in patterns.values()):
            return {}

        # extract the first pattern
        (auxvar, pattern), *rest = patterns.items()
        matches = {}

        # match the first pattern and create the initial substitutions
        rel = self._etables[pattern.head]
        for enode, args in rel.items():
            if (subst := self._match_args(args, pattern.args)) is not None:
                subst[auxvar.name] = enode
                matches[enode] = subst

        # match the rest of the patterns and extend the substitutions
        for auxvar, pattern in rest:
            rel = self._etables[pattern.head]
            tmp = {}
            for enode, subst in matches.items():
                if args := rel.get(subst[auxvar.name]):
                    if (newsubst := self._match_args(args, pattern.args)) is not None:
                        tmp[enode] = {**subst, **newsubst}
            matches = tmp

        return matches

    def apply(self, rewrites: list[Rewrite]) -> int:
        """Apply the given rewrites to the egraph.

        Iteratively match the patterns and apply the rewrites to the graph. The returned
        number of changes is the number of eclasses that were merged. This is the
        number of changes made to the egraph. The egraph is saturated if the number of
        changes is zero.

        Parameters
        ----------
        rewrites :
            A list of rewrites to apply.

        Returns
        -------
        n_changes
            The number of changes made to the egraph.

        """
        n_changes = 0
        for rewrite in promote_list(rewrites):
            for match, subst in self.match(rewrite.matcher).items():
                enode = rewrite.applier.substitute(self, match, subst)
                enode = self.add(enode)
                n_changes += self._eclasses.union(match, enode)
        return n_changes

    def run(self, rewrites: list[Rewrite], n: int = 10) -> bool:
        """Run the match-apply cycles for the given number of iterations.

        Parameters
        ----------
        rewrites :
            A list of rewrites to apply.
        n :
            The number of iterations to run.

        Returns
        -------
        saturated :
            True if the egraph is saturated, False otherwise.

        """
        return any(not self.apply(rewrites) for _i in range(n))

    # TODO(kszucs): investigate whether the costs and best enodes could be maintained
    # during the union operations after each match-apply cycle
    def extract(self, node: Node) -> Node:
        """Extract a node from the egraph.

        The node is converted to an enode which recursively gets converted to an
        enode having the lowest cost according to equivalence classes. Currently
        the cost function is hardcoded as the depth of the enode.

        Parameters
        ----------
        node :
            The node to extract from the egraph.

        Returns
        -------
        node :
            The extracted node.

        """
        enode = self._as_enode(node)
        enode = self._eclasses.find(enode)
        costs = {en: (math.inf, None) for en in self._eclasses.keys()}

        def enode_cost(enode):
            cost = 1
            for arg in enode.args:
                if isinstance(arg, ENode):
                    cost += costs[arg][0]
                else:
                    cost += 1
            return cost

        changed = True
        while changed:
            changed = False
            for en, enodes in self._eclasses.items():
                new_cost = min((enode_cost(en), en) for en in enodes)
                if costs[en][0] != new_cost[0]:
                    changed = True
                costs[en] = new_cost

        def extract(en):
            if not isinstance(en, ENode):
                return en
            best = costs[en][1]
            args = tuple(extract(a) for a in best.args)
            return best.head(*args)

        return extract(enode)

    def equivalent(self, node1: Node, node2: Node) -> bool:
        """Check if two nodes are equivalent.

        The nodes are converted to enodes and checked for equivalence: they are
        equivalent if they are in the same equivalence class.

        Parameters
        ----------
        node1 :
            The first node.
        node2 :
            The second node.

        Returns
        -------
        equivalent :
            True if the nodes are equivalent, False otherwise.

        """
        enode1 = self._as_enode(node1)
        enode2 = self._as_enode(node2)
        enode1 = self._eclasses.find(enode1)
        enode2 = self._eclasses.find(enode2)
        return enode1 == enode2



from __future__ import annotations

import datetime

from ibis.formats.pandas import PandasData


class OraclePandasData(PandasData):
    @classmethod
    def convert_Timestamp_element(cls, dtype):
        return datetime.datetime.fromisoformat

    @classmethod
    def convert_Date_element(cls, dtype):
        return datetime.date.fromisoformat

    @classmethod
    def convert_Time_element(cls, dtype):
        return datetime.time.fromisoformat



from __future__ import annotations

import ibis


def test_failed_column_inference(con):
    # This is a table in the Docker container that we know fails
    # column type inference, so if is loaded, then we're in OK shape.
    table = con.table("ALL_DOMAINS", schema="SYS")
    assert len(table.columns)


def test_blob_raw(con):
    con.drop_table("blob_raw_blobs_blob_raw", force=True)

    with con.begin() as bind:
        bind.execute(
            """CREATE TABLE "blob_raw_blobs_blob_raw" ("blob" BLOB, "raw" RAW(255))"""
        )

    raw_blob = con.table("blob_raw_blobs_blob_raw")

    assert raw_blob.schema() == ibis.Schema(dict(blob="binary", raw="binary"))



from __future__ import annotations

import operator
from datetime import date, datetime

import numpy as np
import pandas as pd
import pandas.testing as tm
import pytest

import ibis
import ibis.expr.datatypes as dt
from ibis import literal as L

pytest.importorskip("clickhouse_connect")


@pytest.mark.parametrize(
    ("left", "right", "type"),
    [
        (L("2017-04-01"), date(2017, 4, 2), dt.date),
        (date(2017, 4, 2), L("2017-04-01"), dt.date),
        (
            L("2017-04-01 01:02:33"),
            datetime(2017, 4, 1, 1, 3, 34),
            dt.timestamp,
        ),
        (
            datetime(2017, 4, 1, 1, 3, 34),
            L("2017-04-01 01:02:33"),
            dt.timestamp,
        ),
    ],
)
@pytest.mark.parametrize(
    "op",
    [
        operator.eq,
        operator.ne,
        operator.lt,
        operator.le,
        operator.gt,
        operator.ge,
    ],
)
def test_string_temporal_compare(con, op, left, right, type):
    expr = op(left, right)
    result = con.execute(expr)
    left_raw = con.execute(L(left).cast(type))
    right_raw = con.execute(L(right).cast(type))
    expected = op(left_raw, right_raw)
    assert result == expected


@pytest.mark.parametrize(
    "op",
    ["add", "sub", "mul", "truediv", "pow", "lt", "le", "gt", "ge", "eq", "ne"],
)
def test_binary_infix_operators(con, alltypes, op, snapshot):
    func = getattr(operator, op)
    a, b = alltypes.int_col, alltypes.tinyint_col
    expr = func(a, b)
    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


# TODO: test boolean operators
# (h & bool_col, '`h` AND (`a` > 0)'),
# (h | bool_col, '`h` OR (`a` > 0)'),
# (h ^ bool_col, 'xor(`h`, (`a` > 0))')


@pytest.mark.parametrize(
    "op",
    [
        lambda a, b, c: (a + b) + c,
        lambda a, _, c: a.log() + c,
        lambda a, b, c: (b + (-(a + c))),
    ],
)
def test_binary_infix_parenthesization(con, alltypes, op, snapshot):
    a = alltypes.int_col
    b = alltypes.tinyint_col
    c = alltypes.double_col

    expr = op(a, b, c)
    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


def test_between(con, alltypes, snapshot):
    expr = alltypes.int_col.between(0, 10)
    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


@pytest.mark.parametrize(
    ("left", "right"),
    [
        (L("2017-03-31").cast(dt.date), date(2017, 4, 2)),
        (date(2017, 3, 31), L("2017-04-02").cast(dt.date)),
    ],
)
def test_string_temporal_compare_between_dates(con, left, right):
    expr = ibis.timestamp("2017-04-01").cast(dt.date).between(left, right)
    result = con.execute(expr)
    assert result


@pytest.mark.parametrize(
    ("left", "right"),
    [
        (
            L("2017-03-31 00:02:33").cast(dt.timestamp),
            datetime(2017, 4, 1, 1, 3, 34),
        ),
        (
            datetime(2017, 3, 31, 0, 2, 33),
            L("2017-04-01 01:03:34").cast(dt.timestamp),
        ),
    ],
)
def test_string_temporal_compare_between_datetimes(con, left, right):
    expr = ibis.timestamp("2017-04-01 00:02:34").between(left, right)
    result = con.execute(expr)
    assert result


@pytest.mark.parametrize("container", [list, tuple, set])
def test_field_in_literals(con, alltypes, df, container):
    values = {"1", "2", "3", "5", "7"}
    foobar = container(values)

    expr = alltypes.string_col.isin(foobar)
    result_col = con.execute(expr.name("result"))
    expected_col = df.string_col.isin(foobar).rename("result")
    tm.assert_series_equal(result_col, expected_col)

    expr = alltypes.string_col.notin(foobar)
    result_col = con.execute(expr.name("result"))
    expected_col = ~df.string_col.isin(foobar).rename("result")
    tm.assert_series_equal(result_col, expected_col)


@pytest.mark.parametrize("column", ["int_col", "float_col", "bool_col"])
def test_negate(con, alltypes, column, snapshot):
    expr = -alltypes[column]
    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


@pytest.mark.parametrize(
    "field",
    [
        "tinyint_col",
        "smallint_col",
        "int_col",
        "bigint_col",
        "float_col",
        "double_col",
        "year",
        "month",
    ],
)
def test_negate_non_boolean(alltypes, field, df):
    t = alltypes.limit(10)
    expr = t.select((-t[field]).name(field))
    result = expr.execute()[field]
    expected = -df.head(10)[field]
    tm.assert_series_equal(result, expected)


def test_negate_literal(con):
    expr = -L(5.245)
    assert round(con.execute(expr), 3) == -5.245


@pytest.mark.parametrize(
    ("op", "pandas_op"),
    [
        (
            lambda t: (t.double_col > 20).ifelse(10, -20),
            lambda df: pd.Series(np.where(df.double_col > 20, 10, -20), dtype="int8"),
        ),
        (
            lambda t: (t.double_col > 20).ifelse(10, -20).abs(),
            lambda df: (
                pd.Series(np.where(df.double_col > 20, 10, -20)).abs().astype("int8")
            ),
        ),
    ],
)
def test_ifelse(alltypes, df, op, pandas_op):
    expr = op(alltypes)
    result = expr.execute()
    result.name = None
    expected = pandas_op(df)

    tm.assert_series_equal(result, expected)


def test_simple_case(con, alltypes, snapshot):
    t = alltypes
    expr = (
        t.string_col.case().when("foo", "bar").when("baz", "qux").else_("default").end()
    )

    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


def test_search_case(con, alltypes, snapshot):
    t = alltypes
    expr = (
        ibis.case()
        .when(t.float_col > 0, t.int_col * 2)
        .when(t.float_col < 0, t.int_col)
        .else_(0)
        .end()
    )

    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


@pytest.mark.parametrize(
    "arr",
    [
        [1, 2, 3],
        ["qw", "wq", "1"],
        [1.2, 0.3, 0.4],
        [[1], [1, 2], [1, 2, 3]],
    ],
)
@pytest.mark.parametrize(
    "gen_idx",
    [
        lambda arr: range(len(arr)),
        lambda arr: range(-len(arr), 0),
    ],
    ids=["positive", "negative"],
)
def test_array_index(con, arr, gen_idx):
    expr = L(arr)
    for i in gen_idx(arr):
        el_expr = expr[i]
        el = con.execute(el_expr)
        assert el == arr[i]


@pytest.mark.parametrize(
    "arrays",
    [
        ([1], [2]),
        ([1], [1, 2]),
        ([1, 2], [1]),
        ([1, 2], [3, 4]),
        ([1, 2], [3, 4], [5, 6]),
    ],
)
def test_array_concat(con, arrays):
    expr = L([]).cast("!array<int8>")
    expected = sum(arrays, [])  # noqa: RUF017
    for arr in arrays:
        expr += L(arr, type="!array<int8>")

    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ("arr", "times"),
    [([1], 1), ([1], 2), ([1], 3), ([1, 2], 1), ([1, 2], 2), ([1, 2], 3)],
)
def test_array_repeat(con, arr, times):
    expected = arr * times
    expr = L(arr) * times
    assert con.execute(expr) == expected


@pytest.mark.parametrize("arr", [[], [1], [1, 2, 3, 4, 5, 6]])
@pytest.mark.parametrize("start", [None, 0, 1, 2, -1, -3])
@pytest.mark.parametrize("stop", [None, 0, 1, 3, -2, -4])
def test_array_slice(con, arr, start, stop):
    expr = L(arr, type="array<int8>")
    assert con.execute(expr[start:stop]) == arr[start:stop]
